from __future__ import annotations

import enum
import reprlib
from collections import defaultdict
from collections.abc import AsyncIterator, Callable, Iterable, Iterator, Sequence, Sized
from collections.abc import Set as AbstractSet
from contextlib import asynccontextmanager, contextmanager
from dataclasses import dataclass
from enum import auto
from functools import reduce
from math import floor
from operator import ge, itemgetter, le, or_
from re import search
from typing import TYPE_CHECKING, Any, Literal, TypeGuard, assert_never, cast, overload

import sqlalchemy
from sqlalchemy import (
    URL,
    Boolean,
    Column,
    Connection,
    DateTime,
    Engine,
    Float,
    Insert,
    Interval,
    LargeBinary,
    MetaData,
    Numeric,
    PrimaryKeyConstraint,
    Selectable,
    String,
    Table,
    Unicode,
    UnicodeText,
    Uuid,
    and_,
    case,
    insert,
    quoted_name,
    text,
)
from sqlalchemy import create_engine as _create_engine
from sqlalchemy.dialects.mssql import dialect as mssql_dialect
from sqlalchemy.dialects.mysql import dialect as mysql_dialect
from sqlalchemy.dialects.oracle import dialect as oracle_dialect
from sqlalchemy.dialects.postgresql import Insert as postgresql_Insert
from sqlalchemy.dialects.postgresql import dialect as postgresql_dialect
from sqlalchemy.dialects.postgresql import insert as postgresql_insert
from sqlalchemy.dialects.postgresql.asyncpg import PGDialect_asyncpg
from sqlalchemy.dialects.sqlite import Insert as sqlite_Insert
from sqlalchemy.dialects.sqlite import dialect as sqlite_dialect
from sqlalchemy.dialects.sqlite import insert as sqlite_insert
from sqlalchemy.exc import ArgumentError, DatabaseError
from sqlalchemy.ext.asyncio import AsyncConnection, AsyncEngine, create_async_engine
from sqlalchemy.orm import (
    DeclarativeBase,
    InstrumentedAttribute,
    class_mapper,
    declared_attr,
)
from sqlalchemy.orm.exc import UnmappedClassError
from sqlalchemy.pool import NullPool, Pool
from sqlalchemy.sql.functions import now
from sqlalchemy.sql.schema import ColumnElementColumnDefault
from typing_extensions import override

from utilities.asyncio import timeout_dur
from utilities.datetime import get_now
from utilities.functions import get_class_name
from utilities.iterables import (
    CheckLengthError,
    MaybeIterable,
    OneEmptyError,
    always_iterable,
    check_length,
    chunked,
    is_iterable_not_str,
    one,
)
from utilities.text import ensure_str
from utilities.types import (
    Duration,
    StrMapping,
    TupleOrStrMapping,
    is_string_mapping,
    is_tuple_or_string_mapping,
)

if TYPE_CHECKING:
    from sqlalchemy.sql.base import ReadOnlyColumnCollection


EngineOrConnection = Engine | Connection
AsyncEngineOrConnection = AsyncEngine | AsyncConnection
MaybeAsyncEngineOrConnection = EngineOrConnection | AsyncEngineOrConnection
TableOrMappedClass = Table | type[DeclarativeBase]
CHUNK_SIZE_FRAC = 0.95


def _check_column_collections_equal(
    x: ReadOnlyColumnCollection[Any, Any],
    y: ReadOnlyColumnCollection[Any, Any],
    /,
    *,
    snake: bool = False,
    allow_permutations: bool = False,
    primary_key: bool = True,
) -> None:
    """Check that a pair of column collections are equal."""
    from utilities.humps import snake_case_mappings

    cols_x, cols_y = (list(cast(Iterable[Column[Any]], i)) for i in [x, y])
    name_to_col_x, name_to_col_y = (
        {ensure_str(col.name): col for col in i} for i in [cols_x, cols_y]
    )
    if len(name_to_col_x) != len(name_to_col_y):
        msg = f"{x=}, {y=}"
        raise _CheckColumnCollectionsEqualError(msg)
    if snake:
        name_to_snake_x, name_to_snake_y = (
            snake_case_mappings(i) for i in [name_to_col_x, name_to_col_y]
        )
        snake_to_name_x, snake_to_name_y = (
            {v: k for k, v in nts.items()} for nts in [name_to_snake_x, name_to_snake_y]
        )
        key_to_col_x, key_to_col_y = (
            {key: name_to_col[snake_to_name[key]] for key in snake_to_name}
            for name_to_col, snake_to_name in [
                (name_to_col_x, snake_to_name_x),
                (name_to_col_y, snake_to_name_y),
            ]
        )
    else:
        key_to_col_x, key_to_col_y = name_to_col_x, name_to_col_y
    if allow_permutations:
        cols_to_check_x, cols_to_check_y = (
            map(itemgetter(1), sorted(key_to_col.items(), key=itemgetter(0)))
            for key_to_col in [key_to_col_x, key_to_col_y]
        )
    else:
        cols_to_check_x, cols_to_check_y = (
            i.values() for i in [key_to_col_x, key_to_col_y]
        )
    diff = set(key_to_col_x).symmetric_difference(set(key_to_col_y))
    if len(diff) >= 1:
        msg = f"{x=}, {y=}"
        raise _CheckColumnCollectionsEqualError(msg)
    for x_i, y_i in zip(cols_to_check_x, cols_to_check_y, strict=True):
        _check_columns_equal(x_i, y_i, snake=snake, primary_key=primary_key)


class _CheckColumnCollectionsEqualError(Exception): ...


def _check_columns_equal(
    x: Column[Any], y: Column[Any], /, *, snake: bool = False, primary_key: bool = True
) -> None:
    """Check that a pair of columns are equal."""
    _check_table_or_column_names_equal(x.name, y.name, snake=snake)
    _check_column_types_equal(x.type, y.type)
    if primary_key and (x.primary_key != y.primary_key):
        msg = f"{x.primary_key=}, {y.primary_key=}"
        raise _CheckColumnsEqualError(msg)
    if x.nullable != y.nullable:
        msg = f"{x.nullable=}, {y.nullable=}"
        raise _CheckColumnsEqualError(msg)


class _CheckColumnsEqualError(Exception): ...


def _check_column_types_equal(x: Any, y: Any, /) -> None:
    """Check that a pair of column types are equal."""
    x_inst, y_inst = (i() if isinstance(i, type) else i for i in [x, y])
    x_cls, y_cls = (i._type_affinity for i in [x_inst, y_inst])  # noqa: SLF001
    msg = f"{x=}, {y=}"
    if not (isinstance(x_inst, y_cls) and isinstance(y_inst, x_cls)):
        raise _CheckColumnTypesEqualError(msg)
    if isinstance(x_inst, Boolean) and isinstance(y_inst, Boolean):
        _check_column_types_boolean_equal(x_inst, y_inst)
    if isinstance(x_inst, DateTime) and isinstance(y_inst, DateTime):
        _check_column_types_datetime_equal(x_inst, y_inst)
    if isinstance(x_inst, sqlalchemy.Enum) and isinstance(y_inst, sqlalchemy.Enum):
        _check_column_types_enum_equal(x_inst, y_inst)
    if isinstance(x_inst, Float) and isinstance(y_inst, Float):
        _check_column_types_float_equal(x_inst, y_inst)
    if isinstance(x_inst, Interval) and isinstance(y_inst, Interval):
        _check_column_types_interval_equal(x_inst, y_inst)
    if isinstance(x_inst, LargeBinary) and isinstance(y_inst, LargeBinary):
        _check_column_types_large_binary_equal(x_inst, y_inst)
    if isinstance(x_inst, Numeric) and isinstance(y_inst, Numeric):
        _check_column_types_numeric_equal(x_inst, y_inst)
    if isinstance(x_inst, String | Unicode | UnicodeText) and isinstance(
        y_inst, String | Unicode | UnicodeText
    ):
        _check_column_types_string_equal(x_inst, y_inst)
    if isinstance(x_inst, Uuid) and isinstance(y_inst, Uuid):
        _check_column_types_uuid_equal(x_inst, y_inst)


class _CheckColumnTypesEqualError(Exception): ...


def _check_column_types_boolean_equal(x: Any, y: Any, /) -> None:
    """Check that a pair of boolean column types are equal."""
    msg = f"{x=}, {y=}"
    if x.create_constraint is not y.create_constraint:
        raise _CheckColumnTypesBooleanEqualError(msg)
    if x.name != y.name:
        raise _CheckColumnTypesBooleanEqualError(msg)


class _CheckColumnTypesBooleanEqualError(Exception): ...


def _check_column_types_datetime_equal(x: Any, y: Any, /) -> None:
    """Check that a pair of datetime column types are equal."""
    if x.timezone is not y.timezone:
        msg = f"{x=}, {y=}"
        raise _CheckColumnTypesDateTimeEqualError(msg)


class _CheckColumnTypesDateTimeEqualError(Exception): ...


def _check_column_types_enum_equal(x: Any, y: Any, /) -> None:
    """Check that a pair of enum column types are equal."""
    x_enum, y_enum = (i.enum_class for i in [x, y])
    if (x_enum is None) and (y_enum is None):
        return
    msg = f"{x=}, {y=}"
    if ((x_enum is None) and (y_enum is not None)) or (
        (x_enum is not None) and (y_enum is None)
    ):
        raise _CheckColumnTypesEnumEqualError(msg)
    if not (issubclass(x_enum, y_enum) and issubclass(y_enum, x_enum)):
        raise _CheckColumnTypesEnumEqualError(msg)
    if x.create_constraint is not y.create_constraint:
        raise _CheckColumnTypesEnumEqualError(msg)
    if x.native_enum is not y.native_enum:
        raise _CheckColumnTypesEnumEqualError(msg)
    if x.length != y.length:
        raise _CheckColumnTypesEnumEqualError(msg)
    if x.inherit_schema is not y.inherit_schema:
        raise _CheckColumnTypesEnumEqualError(msg)


class _CheckColumnTypesEnumEqualError(Exception): ...


def _check_column_types_float_equal(x: Any, y: Any, /) -> None:
    """Check that a pair of float column types are equal."""
    msg = f"{x=}, {y=}"
    if x.precision != y.precision:
        raise _CheckColumnTypesFloatEqualError(msg)
    if x.asdecimal is not y.asdecimal:
        raise _CheckColumnTypesFloatEqualError(msg)
    if x.decimal_return_scale != y.decimal_return_scale:
        raise _CheckColumnTypesFloatEqualError(msg)


class _CheckColumnTypesFloatEqualError(Exception): ...


def _check_column_types_interval_equal(x: Any, y: Any, /) -> None:
    """Check that a pair of interval column types are equal."""
    msg = f"{x=}, {y=}"
    if x.native is not y.native:
        raise _CheckColumnTypesIntervalEqualError(msg)
    if x.second_precision != y.second_precision:
        raise _CheckColumnTypesIntervalEqualError(msg)
    if x.day_precision != y.day_precision:
        raise _CheckColumnTypesIntervalEqualError(msg)


class _CheckColumnTypesIntervalEqualError(Exception): ...


def _check_column_types_large_binary_equal(x: Any, y: Any, /) -> None:
    """Check that a pair of large binary column types are equal."""
    if x.length != y.length:
        msg = f"{x=}, {y=}"
        raise _CheckColumnTypesLargeBinaryEqualError(msg)


class _CheckColumnTypesLargeBinaryEqualError(Exception): ...


def _check_column_types_numeric_equal(x: Any, y: Any, /) -> None:
    """Check that a pair of numeric column types are equal."""
    msg = f"{x=}, {y=}"
    if x.precision != y.precision:
        raise _CheckColumnTypesNumericEqualError(msg)
    if x.scale != y.scale:
        raise _CheckColumnTypesNumericEqualError(msg)
    if x.asdecimal != y.asdecimal:
        raise _CheckColumnTypesNumericEqualError(msg)
    if x.decimal_return_scale != y.decimal_return_scale:
        raise _CheckColumnTypesNumericEqualError(msg)


class _CheckColumnTypesNumericEqualError(Exception): ...


def _check_column_types_string_equal(x: Any, y: Any, /) -> None:
    """Check that a pair of string column types are equal."""
    msg = f"{x=}, {y=}"
    if x.length != y.length:
        raise _CheckColumnTypesStringEqualError(msg)
    if x.collation != y.collation:
        raise _CheckColumnTypesStringEqualError(msg)


class _CheckColumnTypesStringEqualError(Exception): ...


def _check_column_types_uuid_equal(x: Any, y: Any, /) -> None:
    """Check that a pair of UUID column types are equal."""
    msg = f"{x=}, {y=}"
    if x.as_uuid is not y.as_uuid:
        raise _CheckColumnTypesUuidEqualError(msg)
    if x.native_uuid is not y.native_uuid:
        raise _CheckColumnTypesUuidEqualError(msg)


class _CheckColumnTypesUuidEqualError(Exception): ...


def check_engine(
    engine_or_conn: EngineOrConnection,
    /,
    *,
    num_tables: int | tuple[int, float] | None = None,
) -> None:
    """Check that an engine can connect.

    Optionally query for the number of tables, or the number of columns in
    such a table.
    """
    match get_dialect(engine_or_conn):
        case (  # skipif-ci-and-not-linux
            Dialect.mssql | Dialect.mysql | Dialect.postgresql
        ):
            query = "select * from information_schema.tables"
        case Dialect.oracle:  # pragma: no cover
            query = "select * from all_objects"
        case Dialect.sqlite:
            query = "select * from sqlite_master where type='table'"
        case _ as never:  # pyright: ignore[reportUnnecessaryComparison]
            assert_never(never)
    statement = text(query)
    with yield_connection(engine_or_conn) as conn:
        rows = conn.execute(statement).all()
    if num_tables is not None:
        try:
            check_length(rows, equal_or_approx=num_tables)
        except CheckLengthError as error:
            raise CheckEngineError(
                engine_or_conn=engine_or_conn, rows=error.obj, expected=num_tables
            ) from None


@dataclass(kw_only=True, slots=True)
class CheckEngineError(Exception):
    engine_or_conn: EngineOrConnection
    rows: Sized
    expected: int | tuple[int, float]

    @override
    def __str__(self) -> str:
        return f"{reprlib.repr(self.engine_or_conn)} must have {self.expected} table(s); got {len(self.rows)}"


def check_table_against_reflection(
    table_or_mapped_class: TableOrMappedClass,
    engine_or_conn: EngineOrConnection,
    /,
    *,
    schema: str | None = None,
    snake_table: bool = False,
    snake_columns: bool = False,
    allow_permutations_columns: bool = False,
    primary_key: bool = True,
) -> None:
    """Check that a table equals its reflection."""
    reflected = reflect_table(table_or_mapped_class, engine_or_conn, schema=schema)
    _check_tables_equal(
        reflected,
        table_or_mapped_class,
        snake_table=snake_table,
        allow_permutations_columns=allow_permutations_columns,
        snake_columns=snake_columns,
        primary_key=primary_key,
    )


def _check_tables_equal(
    x: TableOrMappedClass,
    y: TableOrMappedClass,
    /,
    *,
    snake_table: bool = False,
    snake_columns: bool = False,
    allow_permutations_columns: bool = False,
    primary_key: bool = True,
) -> None:
    """Check that a pair of tables are equal."""
    x_t, y_t = map(get_table, [x, y])
    _check_table_or_column_names_equal(x_t.name, y_t.name, snake=snake_table)
    _check_column_collections_equal(
        x_t.columns,
        y_t.columns,
        snake=snake_columns,
        allow_permutations=allow_permutations_columns,
        primary_key=primary_key,
    )


def _check_table_or_column_names_equal(
    x: str | quoted_name, y: str | quoted_name, /, *, snake: bool = False
) -> None:
    """Check that a pair of table/columns' names are equal."""
    from utilities.humps import snake_case

    x, y = (str(i) if isinstance(i, quoted_name) else i for i in [x, y])
    msg = f"{x=}, {y=}"
    if (not snake) and (x != y):
        raise _CheckTableOrColumnNamesEqualError(msg)
    if snake and (snake_case(x) != snake_case(y)):
        raise _CheckTableOrColumnNamesEqualError(msg)


class _CheckTableOrColumnNamesEqualError(Exception): ...


def columnwise_max(*columns: Any) -> Any:
    """Compute the columnwise max of a number of columns."""
    return _columnwise_minmax(*columns, op=ge)


def columnwise_min(*columns: Any) -> Any:
    """Compute the columnwise min of a number of columns."""
    return _columnwise_minmax(*columns, op=le)


def _columnwise_minmax(*columns: Any, op: Callable[[Any, Any], Any]) -> Any:
    """Compute the columnwise min of a number of columns."""

    def func(x: Any, y: Any, /) -> Any:
        x_none = x.is_(None)
        y_none = y.is_(None)
        col = case(
            (and_(x_none, y_none), None),
            (and_(~x_none, y_none), x),
            (and_(x_none, ~y_none), y),
            (op(x, y), x),
            else_=y,
        )
        # try auto-label
        names = {
            value for col in [x, y] if (value := getattr(col, "name", None)) is not None
        }
        try:
            (name,) = names
        except ValueError:
            return col
        else:
            return col.label(name)

    return reduce(func, columns)


@overload
def create_engine(
    drivername: str,
    /,
    *,
    username: str | None = ...,
    password: str | None = ...,
    host: str | None = ...,
    port: int | None = ...,
    database: str | None = ...,
    query: StrMapping | None = ...,
    poolclass: type[Pool] | None = ...,
    async_: Literal[True],
) -> AsyncEngine: ...
@overload
def create_engine(
    drivername: str,
    /,
    *,
    username: str | None = ...,
    password: str | None = ...,
    host: str | None = ...,
    port: int | None = ...,
    database: str | None = ...,
    query: StrMapping | None = ...,
    poolclass: type[Pool] | None = ...,
    async_: Literal[False] = False,
) -> Engine: ...
@overload
def create_engine(
    drivername: str,
    /,
    *,
    username: str | None = ...,
    password: str | None = ...,
    host: str | None = ...,
    port: int | None = ...,
    database: str | None = ...,
    query: StrMapping | None = ...,
    poolclass: type[Pool] | None = ...,
    async_: bool = False,
) -> Engine | AsyncEngine: ...
def create_engine(
    drivername: str,
    /,
    *,
    username: str | None = None,
    password: str | None = None,
    host: str | None = None,
    port: int | None = None,
    database: str | None = None,
    query: StrMapping | None = None,
    poolclass: type[Pool] | None = NullPool,
    async_: bool = False,
) -> Engine | AsyncEngine:
    """Create a SQLAlchemy engine."""
    if query is None:
        kwargs = {}
    else:

        def func(x: MaybeIterable[str], /) -> list[str] | str:
            return x if isinstance(x, str) else list(x)

        kwargs = {"query": {k: func(v) for k, v in query.items()}}
    url = URL.create(
        drivername,
        username=username,
        password=password,
        host=host,
        port=port,
        database=database,
        **kwargs,
    )
    if async_:
        return create_async_engine(url, poolclass=poolclass)
    return _create_engine(url, poolclass=poolclass)


class Dialect(enum.Enum):
    """An enumeration of the SQL dialects."""

    mssql = auto()
    mysql = auto()
    oracle = auto()
    postgresql = auto()
    sqlite = auto()

    @property
    def max_params(self, /) -> int:
        match self:
            case Dialect.mssql:  # pragma: no cover
                return 2100
            case Dialect.mysql:  # pragma: no cover
                return 65535
            case Dialect.oracle:  # pragma: no cover
                return 1000
            case Dialect.postgresql:  # skipif-ci-and-not-linux
                return 32767
            case Dialect.sqlite:
                return 100
            case _ as never:  # pyright: ignore[reportUnnecessaryComparison]
                assert_never(never)


def ensure_engine(engine: Engine | str, /) -> Engine:
    """Ensure the object is an Engine."""
    if isinstance(engine, Engine):
        return engine
    return parse_engine(engine)


def ensure_tables_created(
    engine_or_conn: EngineOrConnection, /, *tables_or_mapped_classes: TableOrMappedClass
) -> None:
    """Ensure a table/set of tables is/are created."""
    prepared = _ensure_tables_created_prepare(engine_or_conn, *tables_or_mapped_classes)
    for table in prepared.tables:
        with yield_connection(engine_or_conn) as conn:
            try:
                table.create(conn)
            except DatabaseError as error:
                _ensure_tables_maybe_reraise(error, prepared.match)


async def ensure_tables_created_async(
    engine_or_conn: AsyncEngineOrConnection,
    /,
    *tables_or_mapped_classes: TableOrMappedClass,
    timeout: Duration | None = None,  # noqa: ASYNC109
) -> None:
    """Ensure a table/set of tables is/are created."""
    prepared = _ensure_tables_created_prepare(engine_or_conn, *tables_or_mapped_classes)
    for table in prepared.tables:
        async with yield_connection_async(engine_or_conn, timeout=timeout) as conn:
            try:
                await conn.run_sync(table.create)
            except DatabaseError as error:
                _ensure_tables_maybe_reraise(error, prepared.match)


@dataclass(kw_only=True, slots=True)
class _EnsureTablesCreatedOrDroppedPrepare:
    match: str
    tables: AbstractSet[Table]


def _ensure_tables_created_prepare(
    engine_or_conn: MaybeAsyncEngineOrConnection,
    /,
    *tables_or_mapped_classes: TableOrMappedClass,
) -> _EnsureTablesCreatedOrDroppedPrepare:
    """Prepare the arguments for `ensure_tables_created`."""
    return _EnsureTablesCreatedOrDroppedPrepare(
        match=_ensure_tables_created_match(engine_or_conn),
        tables=set(map(get_table, tables_or_mapped_classes)),
    )


def _ensure_tables_created_match(
    engine_or_conn: MaybeAsyncEngineOrConnection, /
) -> str:
    """Get the match statement for the given engine."""
    match dialect := get_dialect(engine_or_conn):
        case Dialect.mysql:  # pragma: no cover
            raise NotImplementedError(dialect)
        case Dialect.postgresql:  # skipif-ci-and-not-linux
            return "relation .* already exists"
        case Dialect.mssql:  # pragma: no cover
            return "There is already an object named .* in the database"
        case Dialect.oracle:  # pragma: no cover
            return "ORA-00955: name is already used by an existing object"
        case Dialect.sqlite:
            return "table .* already exists"
        case _ as never:  # pyright: ignore[reportUnnecessaryComparison]
            assert_never(never)


def _ensure_tables_maybe_reraise(error: DatabaseError, match: str, /) -> None:
    """Re-raise the error if it does not match the required statement."""
    if not search(match, ensure_str(one(error.args))):
        raise error  # pragma: no cover


def ensure_tables_dropped(
    engine_or_conn: EngineOrConnection, *tables_or_mapped_classes: TableOrMappedClass
) -> None:
    """Ensure a table/set of tables is/are dropped."""
    prepared = _ensure_tables_dropped_prepare(engine_or_conn, *tables_or_mapped_classes)
    for table in prepared.tables:
        with yield_connection(engine_or_conn) as conn:
            try:
                table.drop(conn)
            except DatabaseError as error:
                _ensure_tables_dropped_maybe_reraise(error, prepared.match)


async def ensure_tables_dropped_async(
    engine_or_conn: AsyncEngineOrConnection,
    *tables_or_mapped_classes: TableOrMappedClass,
    timeout: Duration | None = None,  # noqa: ASYNC109
) -> None:
    """Ensure a table/set of tables is/are dropped."""
    prepared = _ensure_tables_dropped_prepare(engine_or_conn, *tables_or_mapped_classes)
    for table in prepared.tables:
        async with yield_connection_async(engine_or_conn, timeout=timeout) as conn:
            try:
                await conn.run_sync(table.drop)
            except DatabaseError as error:
                _ensure_tables_maybe_reraise(error, prepared.match)


def _ensure_tables_dropped_prepare(
    engine_or_conn: MaybeAsyncEngineOrConnection,
    /,
    *tables_or_mapped_classes: TableOrMappedClass,
) -> _EnsureTablesCreatedOrDroppedPrepare:
    """Prepare the arguments for `ensure_tables_dropped`."""
    return _EnsureTablesCreatedOrDroppedPrepare(
        match=_ensure_tables_dropped_match(engine_or_conn),
        tables=set(map(get_table, tables_or_mapped_classes)),
    )


def _ensure_tables_dropped_match(
    engine_or_conn: MaybeAsyncEngineOrConnection, /
) -> str:
    """Get the match statement for the given engine."""
    match dialect := get_dialect(engine_or_conn):
        case Dialect.mysql:  # pragma: no cover
            raise NotImplementedError(dialect)
        case Dialect.postgresql:  # skipif-ci-and-not-linux
            return "table .* does not exist"
        case Dialect.mssql:  # pragma: no cover
            return "Cannot drop the table .*, because it does not exist or you do not have permission"
        case Dialect.oracle:  # pragma: no cover
            return "ORA-00942: table or view does not exist"
        case Dialect.sqlite:
            return "no such table"
        case _ as never:  # pyright: ignore[reportUnnecessaryComparison]
            assert_never(never)


def _ensure_tables_dropped_maybe_reraise(error: DatabaseError, match: str, /) -> None:
    """Re-raise the error if it does not match the required statement."""
    if not search(match, ensure_str(one(error.args))):
        raise error  # pragma: no cover


def get_chunk_size(
    engine_or_conn: MaybeAsyncEngineOrConnection,
    /,
    *,
    chunk_size_frac: float = CHUNK_SIZE_FRAC,
    scaling: float = 1.0,
) -> int:
    """Get the maximum chunk size for an engine."""
    dialect = get_dialect(engine_or_conn)
    max_params = dialect.max_params
    return max(floor(chunk_size_frac * max_params / scaling), 1)


def get_column_names(table_or_mapped_class: TableOrMappedClass, /) -> list[str]:
    """Get the column names from a table or model."""
    return [col.name for col in get_columns(table_or_mapped_class)]


def get_columns(table_or_mapped_class: TableOrMappedClass, /) -> list[Column[Any]]:
    """Get the columns from a table or model."""
    return list(get_table(table_or_mapped_class).columns)


def get_dialect(engine_or_conn: MaybeAsyncEngineOrConnection, /) -> Dialect:
    """Get the dialect of a database."""
    dialect = engine_or_conn.dialect
    if isinstance(dialect, mssql_dialect):  # pragma: no cover
        return Dialect.mssql
    if isinstance(dialect, mysql_dialect):  # pragma: no cover
        return Dialect.mysql
    if isinstance(dialect, oracle_dialect):  # pragma: no cover
        return Dialect.oracle
    if isinstance(  # skipif-ci-and-not-linux
        dialect, postgresql_dialect | PGDialect_asyncpg
    ):
        return Dialect.postgresql
    if isinstance(dialect, sqlite_dialect):
        return Dialect.sqlite
    raise GetDialectError(dialect=dialect)  # pragma: no cover


@dataclass(kw_only=True, slots=True)
class GetDialectError(Exception):
    dialect: sqlalchemy.Dialect

    @override
    def __str__(self) -> str:
        return (  # pragma: no cover
            f"Dialect must be one of MS SQL, MySQL, Oracle, PostgreSQL or SQLite; got {self.dialect} instead"
        )


def get_table(obj: TableOrMappedClass, /) -> Table:
    """Get the table from a Table or mapped class."""
    if isinstance(obj, Table):
        return obj
    if is_mapped_class(obj):
        return cast(Table, obj.__table__)
    raise GetTableError(obj=obj)


@dataclass(kw_only=True, slots=True)
class GetTableError(Exception):
    obj: Any

    @override
    def __str__(self) -> str:
        return f"Object {self.obj} must be a Table or mapped class; got {get_class_name(self.obj)!r}"


def get_table_updated_column(
    table_or_mapped_class: TableOrMappedClass, /, *, pattern: str = "updated"
) -> str | None:
    """Get the name of the unique `updated_at` column, if it exists."""

    def is_updated_at(column: Column[Any], /) -> bool:
        return (
            bool(search(pattern, column.name))
            and is_date_time_with_time_zone(column.type)
            and is_now(column.onupdate)
        )

    def is_date_time_with_time_zone(type_: Any, /) -> bool:
        return isinstance(type_, DateTime) and type_.timezone

    def is_now(on_update: Any, /) -> bool:
        return isinstance(on_update, ColumnElementColumnDefault) and isinstance(
            on_update.arg, now
        )

    matches = filter(is_updated_at, get_columns(table_or_mapped_class))
    try:
        return one(matches).name
    except OneEmptyError:
        return None


def get_table_name(table_or_mapped_class: TableOrMappedClass, /) -> str:
    """Get the table name from a Table or mapped class."""
    return get_table(table_or_mapped_class).name


_PairOfTupleAndTable = tuple[tuple[Any, ...], TableOrMappedClass]
_PairOfDictAndTable = tuple[StrMapping, TableOrMappedClass]
_PairOfListOfTuplesAndTable = tuple[Sequence[tuple[Any, ...]], TableOrMappedClass]
_PairOfListOfDictsAndTable = tuple[Sequence[StrMapping], TableOrMappedClass]
_ListOfPairOfTupleAndTable = Sequence[tuple[tuple[Any, ...], TableOrMappedClass]]
_ListOfPairOfDictAndTable = Sequence[tuple[StrMapping, TableOrMappedClass]]
_InsertItem = (
    _PairOfTupleAndTable
    | _PairOfDictAndTable
    | _PairOfListOfTuplesAndTable
    | _PairOfListOfDictsAndTable
    | _ListOfPairOfTupleAndTable
    | _ListOfPairOfDictAndTable
    | MaybeIterable[DeclarativeBase]
)


def insert_items(
    engine_or_conn: EngineOrConnection,
    /,
    *items: _InsertItem,
    chunk_size_frac: float = CHUNK_SIZE_FRAC,
    assume_tables_exist: bool = False,
) -> None:
    """Insert a set of items into a database.

    These can be one of the following:
     - pair of tuple & table/class:           (x1, x2, ...), table_cls
     - pair of dict & table/class:            {k1=v1, k2=v2, ...), table_cls
     - pair of list of tuples & table/class:  [(x11, x12, ...),
                                               (x21, x22, ...),
                                               ...], table_cls
     - pair of list of dicts & table/class:   [{k1=v11, k2=v12, ...},
                                               {k1=v21, k2=v22, ...},
                                               ...], table/class
     - list of pairs of tuple & table/class:  [((x11, x12, ...), table_cls1),
                                               ((x21, x22, ...), table_cls2),
                                               ...]
     - list of pairs of dict & table/class:   [({k1=v11, k2=v12, ...}, table_cls1),
                                               ({k1=v21, k2=v22, ...}, table_cls2),
                                               ...]
     - mapped class:                          Obj(k1=v1, k2=v2, ...)
     - list of mapped classes:                [Obj(k1=v11, k2=v12, ...),
                                               Obj(k1=v21, k2=v22, ...),
                                               ...]
    """
    try:
        prepared = _insert_items_prepare(
            engine_or_conn, *items, chunk_size_frac=chunk_size_frac
        )
    except _InsertItemsPrepareError as error:
        raise InsertItemsError(item=error.item) from None
    if not assume_tables_exist:
        ensure_tables_created(engine_or_conn, *prepared.tables)
    for ins, parameters in prepared.yield_pairs():
        with yield_connection(engine_or_conn) as conn:
            _ = conn.execute(ins, parameters=parameters)


@dataclass(kw_only=True, slots=True)
class InsertItemsError(Exception):
    item: _InsertItem

    @override
    def __str__(self) -> str:
        return f"Item must be valid; got {self.item}"


async def insert_items_async(
    engine_or_conn: AsyncEngineOrConnection,
    *items: _InsertItem,
    chunk_size_frac: float = CHUNK_SIZE_FRAC,
    assume_tables_exist: bool = False,
    timeout: Duration | None = None,  # noqa: ASYNC109
) -> None:
    """Insert a set of items into a database.

    These can be one of the following:
     - pair of tuple & table/class:           (x1, x2, ...), table_cls
     - pair of dict & table/class:            {k1=v1, k2=v2, ...), table_cls
     - pair of list of tuples & table/class:  [(x11, x12, ...),
                                               (x21, x22, ...),
                                               ...], table_cls
     - pair of list of dicts & table/class:   [{k1=v11, k2=v12, ...},
                                               {k1=v21, k2=v22, ...},
                                               ...], table/class
     - list of pairs of tuple & table/class:  [((x11, x12, ...), table_cls1),
                                               ((x21, x22, ...), table_cls2),
                                               ...]
     - list of pairs of dict & table/class:   [({k1=v11, k2=v12, ...}, table_cls1),
                                               ({k1=v21, k2=v22, ...}, table_cls2),
                                               ...]
     - mapped class:                          Obj(k1=v1, k2=v2, ...)
     - list of mapped classes:                [Obj(k1=v11, k2=v12, ...),
                                               Obj(k1=v21, k2=v22, ...),
                                               ...]
    """
    try:
        prepared = _insert_items_prepare(
            engine_or_conn, *items, chunk_size_frac=chunk_size_frac
        )
    except _InsertItemsPrepareError as error:
        raise InsertItemsAsyncError(item=error.item) from None
    if not assume_tables_exist:
        await ensure_tables_created_async(
            engine_or_conn, *prepared.tables, timeout=timeout
        )
    for ins, parameters in prepared.yield_pairs():
        async with yield_connection_async(engine_or_conn, timeout=timeout) as conn:
            _ = await conn.execute(ins, parameters=parameters)


@dataclass(kw_only=True, slots=True)
class InsertItemsAsyncError(Exception):
    item: _InsertItem

    @override
    def __str__(self) -> str:
        return f"Item must be valid; got {self.item}"


@dataclass(kw_only=True, slots=True)
class _PreInsertUpsertItems:
    tables: Sequence[Table]
    yield_pairs: Callable[[], Iterator[tuple[Insert, Any]]]


def _insert_items_prepare(
    engine_or_conn: MaybeAsyncEngineOrConnection,
    /,
    *items: _InsertItem,
    chunk_size_frac: float = CHUNK_SIZE_FRAC,
) -> _PreInsertUpsertItems:
    """Prepare the arguments for `insert_items`."""
    mapping: dict[Table, list[TupleOrStrMapping]] = defaultdict(list)
    lengths: set[int] = set()
    try:
        for item in items:
            for normed in _normalize_insert_item(item):
                values = normed.values
                mapping[normed.table].append(values)
                lengths.add(len(values))
    except _NormalizeInsertItemError as error:
        raise _InsertItemsPrepareError(item=error.item) from None
    tables = list(mapping)
    max_length = max(lengths, default=1)
    chunk_size = get_chunk_size(
        engine_or_conn, chunk_size_frac=chunk_size_frac, scaling=max_length
    )

    def yield_pairs() -> Iterator[tuple[Insert, Any]]:
        for table, values in mapping.items():
            for chunk in chunked(values, chunk_size):
                match get_dialect(engine_or_conn):
                    case Dialect.oracle:  # pragma: no cover
                        yield insert(table), chunk
                    case _:
                        yield insert(table).values(list(chunk)), None

    return _PreInsertUpsertItems(tables=tables, yield_pairs=yield_pairs)


@dataclass(kw_only=True, slots=True)
class _InsertItemsPrepareError(Exception):
    item: _InsertItem

    @override
    def __str__(self) -> str:
        return f"Item must be valid; got {self.item}"


def is_insert_item_pair(
    obj: Any, /
) -> TypeGuard[tuple[TupleOrStrMapping, TableOrMappedClass]]:
    """Check if an object is an insert-ready pair."""
    return _is_insert_or_upsert_pair(obj, is_tuple_or_string_mapping)


def is_upsert_item_pair(
    obj: Any, /
) -> TypeGuard[tuple[StrMapping, TableOrMappedClass]]:
    """Check if an object is an upsert-ready pair."""
    return _is_insert_or_upsert_pair(obj, is_string_mapping)


def _is_insert_or_upsert_pair(
    obj: Any, predicate: Callable[[TupleOrStrMapping], bool], /
) -> bool:
    """Check if an object is an insert/upsert-ready pair."""
    return (
        isinstance(obj, tuple)
        and (len(obj) == 2)
        and predicate(obj[0])
        and is_table_or_mapped_class(obj[1])
    )


def is_mapped_class(obj: Any, /) -> bool:
    """Check if an object is a mapped class."""
    if isinstance(obj, type):
        try:
            _ = class_mapper(cast(Any, obj))
        except (ArgumentError, UnmappedClassError):
            return False
        return True
    return is_mapped_class(type(obj))


def is_table_or_mapped_class(obj: Any, /) -> bool:
    """Check if an object is a Table or a mapped class."""
    return isinstance(obj, Table) or is_mapped_class(obj)


def mapped_class_to_dict(obj: Any, /) -> dict[str, Any]:
    """Construct a dictionary of elements for insertion."""
    cls = type(obj)

    def is_attr(attr: str, key: str, /) -> str | None:
        if isinstance(value := getattr(cls, attr), InstrumentedAttribute) and (
            value.name == key
        ):
            return attr
        return None

    def yield_items() -> Iterator[tuple[str, Any]]:
        for key in get_column_names(cls):
            attr = one(attr for attr in dir(cls) if is_attr(attr, key) is not None)
            yield key, getattr(obj, attr)

    return dict(yield_items())


@dataclass(kw_only=True, slots=True)
class _NormalizedInsertItem:
    values: TupleOrStrMapping
    table: Table


def _normalize_insert_item(item: _InsertItem, /) -> Iterator[_NormalizedInsertItem]:
    """Normalize an insertion item."""
    try:
        for norm in _normalize_upsert_item(cast(Any, item), selected_or_all="all"):
            yield _NormalizedInsertItem(values=norm.values, table=norm.table)
    except _NormalizeUpsertItemError:
        pass
    else:
        return

    if is_insert_item_pair(item):
        yield _NormalizedInsertItem(values=item[0], table=get_table(item[1]))
        return

    item = cast(_PairOfListOfTuplesAndTable | _ListOfPairOfTupleAndTable, item)

    if (
        isinstance(item, tuple)
        and (len(item) == 2)
        and is_iterable_not_str(item[0])
        and all(is_tuple_or_string_mapping(i) for i in item[0])
        and is_table_or_mapped_class(item[1])
    ):
        item = cast(_PairOfListOfTuplesAndTable, item)
        for i in item[0]:
            yield _NormalizedInsertItem(values=i, table=get_table(item[1]))
        return

    item = cast(_ListOfPairOfDictAndTable, item)

    if is_iterable_not_str(item) and all(is_insert_item_pair(i) for i in item):
        item = cast(_ListOfPairOfTupleAndTable | _ListOfPairOfDictAndTable, item)
        for i in item:
            yield _NormalizedInsertItem(values=i[0], table=get_table(i[1]))
        return

    raise _NormalizeInsertItemError(item=item)


@dataclass(kw_only=True, slots=True)
class _NormalizeInsertItemError(Exception):
    item: _InsertItem

    @override
    def __str__(self) -> str:
        return f"Item must be valid; got {self.item}"


@dataclass(kw_only=True, slots=True)
class _NormalizedUpsertItem:
    values: StrMapping
    table: Table


def _normalize_upsert_item(
    item: _UpsertItem, /, *, selected_or_all: Literal["selected", "all"] = "selected"
) -> Iterator[_NormalizedUpsertItem]:
    """Normalize an upsert item."""
    normalized = _normalize_upsert_item_inner(item)
    match selected_or_all:
        case "selected":
            for norm in normalized:
                values = {k: v for k, v in norm.values.items() if v is not None}
                yield _NormalizedUpsertItem(values=values, table=norm.table)
        case "all":
            yield from normalized
        case _ as never:  # pyright: ignore[reportUnnecessaryComparison]
            assert_never(never)


def _normalize_upsert_item_inner(
    item: _UpsertItem, /
) -> Iterator[_NormalizedUpsertItem]:
    if is_upsert_item_pair(item):
        yield _NormalizedUpsertItem(values=item[0], table=get_table(item[1]))
        return

    item = cast(
        _PairOfListOfDictsAndTable
        | _ListOfPairOfDictAndTable
        | DeclarativeBase
        | Sequence[DeclarativeBase],
        item,
    )

    if (
        isinstance(item, tuple)
        and (len(item) == 2)
        and is_iterable_not_str(item[0])
        and all(is_string_mapping(i) for i in item[0])
        and is_table_or_mapped_class(item[1])
    ):
        item = cast(_PairOfListOfDictsAndTable, item)
        for i in item[0]:
            yield _NormalizedUpsertItem(values=i, table=get_table(item[1]))
        return

    item = cast(
        _ListOfPairOfDictAndTable | DeclarativeBase | Sequence[DeclarativeBase], item
    )

    if is_iterable_not_str(item) and all(is_upsert_item_pair(i) for i in item):
        item = cast(_ListOfPairOfDictAndTable, item)
        for i in item:
            yield _NormalizedUpsertItem(values=i[0], table=get_table(i[1]))
        return

    item = cast(MaybeIterable[DeclarativeBase], item)
    if isinstance(item, DeclarativeBase) or (
        is_iterable_not_str(item) and all(isinstance(i, DeclarativeBase) for i in item)
    ):
        for i in always_iterable(item):
            yield _NormalizedUpsertItem(
                values=mapped_class_to_dict(i), table=get_table(i)
            )
        return

    raise _NormalizeUpsertItemError(item=item)


@dataclass(kw_only=True, slots=True)
class _NormalizeUpsertItemError(Exception):
    item: _UpsertItem

    @override
    def __str__(self) -> str:
        return f"Item must be valid; got {self.item}"


def parse_engine(engine: str, /) -> Engine:
    """Parse a string into an Engine."""
    try:
        return _create_engine(engine, poolclass=NullPool)
    except ArgumentError as error:
        raise ParseEngineError(*error.args) from None


class ParseEngineError(Exception): ...


def reflect_table(
    table_or_mapped_class: TableOrMappedClass,
    engine_or_conn: EngineOrConnection,
    /,
    *,
    schema: str | None = None,
) -> Table:
    """Reflect a table from a database."""
    name = get_table_name(table_or_mapped_class)
    metadata = MetaData(schema=schema)
    with yield_connection(engine_or_conn) as conn:
        return Table(name, metadata, autoload_with=conn)


def selectable_to_string(
    selectable: Selectable[Any], engine_or_conn: MaybeAsyncEngineOrConnection, /
) -> str:
    """Convert a selectable into a string."""
    com = selectable.compile(
        dialect=engine_or_conn.dialect, compile_kwargs={"literal_binds": True}
    )
    return str(com)


def serialize_engine(engine: Engine, /) -> str:
    """Serialize an Engine."""
    return engine.url.render_as_string(hide_password=False)


class TablenameMixin:
    """Mix-in for an auto-generated tablename."""

    @cast(Any, declared_attr)
    def __tablename__(cls) -> str:  # noqa: N805
        from utilities.humps import snake_case

        return snake_case(get_class_name(cls))


_UpsertItem = (
    _PairOfDictAndTable
    | _PairOfListOfDictsAndTable
    | _ListOfPairOfDictAndTable
    | MaybeIterable[DeclarativeBase]
)


def upsert_items(
    engine_or_conn: EngineOrConnection,
    /,
    *items: _UpsertItem,
    chunk_size_frac: float = CHUNK_SIZE_FRAC,
    assume_tables_exist: bool = False,
    selected_or_all: Literal["selected", "all"] = "selected",
) -> None:
    """Upsert a set of items into a database.

    These can be one of the following:
     - pair of dict & table/class:            {k1=v1, k2=v2, ...), table_cls
     - pair of list of dicts & table/class:   [{k1=v11, k2=v12, ...},
                                               {k1=v21, k2=v22, ...},
                                               ...], table/class
     - list of pairs of dict & table/class:   [({k1=v11, k2=v12, ...}, table_cls1),
                                               ({k1=v21, k2=v22, ...}, table_cls2),
                                               ...]
     - mapped class:                          Obj(k1=v1, k2=v2, ...)
     - list of mapped classes:                [Obj(k1=v11, k2=v12, ...),
                                               Obj(k1=v21, k2=v22, ...),
                                               ...]
    """
    try:
        prepared = _upsert_items_prepare(
            engine_or_conn,
            *items,
            chunk_size_frac=chunk_size_frac,
            selected_or_all=selected_or_all,
        )
    except _UpsertItemsPrepareError as error:
        raise UpsertItemsError(item=error.item) from None
    if not assume_tables_exist:
        ensure_tables_created(engine_or_conn, *prepared.tables)
    for ins, parameters in prepared.yield_pairs():
        with yield_connection(engine_or_conn) as conn:
            _ = conn.execute(ins, parameters=parameters)


@dataclass(kw_only=True, slots=True)
class UpsertItemsError(Exception):
    item: _InsertItem

    @override
    def __str__(self) -> str:
        return f"Item must be valid; got {self.item}"


def _upsert_items_prepare(
    engine_or_conn: MaybeAsyncEngineOrConnection,
    /,
    *items: _UpsertItem,
    chunk_size_frac: float = CHUNK_SIZE_FRAC,
    selected_or_all: Literal["selected", "all"] = "selected",
) -> _PreInsertUpsertItems:
    """Prepare the arguments for `insert_items`."""
    mapping: dict[Table, list[StrMapping]] = defaultdict(list)
    lengths: set[int] = set()
    try:
        for item in items:
            for normed in _normalize_upsert_item(item, selected_or_all=selected_or_all):
                values = normed.values
                mapping[normed.table].append(values)
                lengths.add(len(values))
    except _NormalizeUpsertItemError as error:
        raise _UpsertItemsPrepareError(item=error.item) from None
    tables = list(mapping)
    max_length = max(lengths, default=1)
    chunk_size = get_chunk_size(
        engine_or_conn, chunk_size_frac=chunk_size_frac, scaling=max_length
    )

    def yield_pairs() -> Iterator[tuple[Insert, Any]]:
        for table, values in mapping.items():
            for chunk in chunked(values, chunk_size):
                ups = _upsert_items_build(
                    engine_or_conn, table, chunk, selected_or_all=selected_or_all
                )
                yield ups, None

    return _PreInsertUpsertItems(tables=tables, yield_pairs=yield_pairs)


def _upsert_items_build(
    engine_or_conn: MaybeAsyncEngineOrConnection,
    table: Table,
    values: Iterable[StrMapping],
    /,
    *,
    selected_or_all: Literal["selected", "all"] = "selected",
) -> Insert:
    values = list(values)
    keys = set(reduce(or_, values))
    dict_nones = {k: None for k in keys}
    values = [{**dict_nones, **v} for v in values]
    if (updated_col := get_table_updated_column(table)) is not None:
        up_map = {updated_col: get_now()}
        values = [{**v, **up_map} for v in values]
    match get_dialect(engine_or_conn):
        case Dialect.postgresql:  # skipif-ci-and-not-linux
            insert = postgresql_insert
        case Dialect.sqlite:
            insert = sqlite_insert
        case (  # pragma: no cover
            (Dialect.mssql | Dialect.mysql | Dialect.oracle) as dialect
        ):
            raise NotImplementedError(dialect)
        case _ as never:  # pyright: ignore[reportUnnecessaryComparison]
            assert_never(never)
    ins = insert(table).values(values)
    primary_key = cast(Any, table.primary_key)
    return _upsert_items_apply_on_conflict_do_update(
        values, ins, primary_key, selected_or_all=selected_or_all
    )


def _upsert_items_apply_on_conflict_do_update(
    values: Iterable[StrMapping],
    insert: postgresql_Insert | sqlite_Insert,
    primary_key: PrimaryKeyConstraint,
    /,
    *,
    selected_or_all: Literal["selected", "all"] = "selected",
) -> Insert:
    match selected_or_all:
        case "selected":
            columns = set(reduce(or_, values))
        case "all":
            columns = {c.name for c in insert.excluded}
        case _ as never:  # pyright: ignore[reportUnnecessaryComparison]
            assert_never(never)
    set_ = {c: getattr(insert.excluded, c) for c in columns}
    match insert:
        case postgresql_Insert():  # skipif-ci
            return insert.on_conflict_do_update(constraint=primary_key, set_=set_)
        case sqlite_Insert():
            return insert.on_conflict_do_update(index_elements=primary_key, set_=set_)
        case _ as never:  # pyright: ignore[reportUnnecessaryComparison]
            assert_never(never)


@dataclass(kw_only=True, slots=True)
class _UpsertItemsPrepareError(Exception):
    item: _InsertItem

    @override
    def __str__(self) -> str:
        return f"Item must be valid; got {self.item}"


async def upsert_items_async(
    engine_or_conn: AsyncEngineOrConnection,
    /,
    *items: _UpsertItem,
    chunk_size_frac: float = CHUNK_SIZE_FRAC,
    selected_or_all: Literal["selected", "all"] = "selected",
    assume_tables_exist: bool = False,
    timeout: Duration | None = None,  # noqa: ASYNC109
) -> None:
    """Upsert a set of items into a database.

    These can be one of the following:
     - pair of dict & table/class:            {k1=v1, k2=v2, ...), table_cls
     - pair of list of dicts & table/class:   [{k1=v11, k2=v12, ...},
                                               {k1=v21, k2=v22, ...},
                                               ...], table/class
     - list of pairs of dict & table/class:   [({k1=v11, k2=v12, ...}, table_cls1),
                                               ({k1=v21, k2=v22, ...}, table_cls2),
                                               ...]
     - mapped class:                          Obj(k1=v1, k2=v2, ...)
     - list of mapped classes:                [Obj(k1=v11, k2=v12, ...),
                                               Obj(k1=v21, k2=v22, ...),
                                               ...]
    """
    try:
        prepared = _upsert_items_prepare(
            engine_or_conn,
            *items,
            chunk_size_frac=chunk_size_frac,
            selected_or_all=selected_or_all,
        )
    except _UpsertItemsPrepareError as error:
        raise UpsertItemsAsyncError(item=error.item) from None
    if not assume_tables_exist:
        await ensure_tables_created_async(
            engine_or_conn, *prepared.tables, timeout=timeout
        )
    for ins, parameters in prepared.yield_pairs():
        async with yield_connection_async(engine_or_conn, timeout=timeout) as conn:
            _ = await conn.execute(ins, parameters=parameters)


@dataclass(kw_only=True, slots=True)
class UpsertItemsAsyncError(Exception):
    item: _InsertItem

    @override
    def __str__(self) -> str:
        return f"Item must be valid; got {self.item}"


@contextmanager
def yield_connection(engine_or_conn: EngineOrConnection, /) -> Iterator[Connection]:
    """Yield a synchronous connection."""
    if isinstance(engine_or_conn, Engine):
        with engine_or_conn.begin() as conn:
            yield conn
    else:
        yield engine_or_conn


@asynccontextmanager
async def yield_connection_async(
    engine_or_conn: AsyncEngineOrConnection,
    /,
    *,
    timeout: Duration | None = None,  # noqa: ASYNC109
) -> AsyncIterator[AsyncConnection]:
    """Yield an asynchronous connection."""
    if isinstance(engine_or_conn, AsyncEngine):
        async with timeout_dur(duration=timeout), engine_or_conn.begin() as conn:
            yield conn
    else:
        yield engine_or_conn


def yield_primary_key_columns(obj: TableOrMappedClass, /) -> Iterator[Column]:
    """Yield the primary key columns of a table."""
    table = get_table(obj)
    yield from table.primary_key


__all__ = [
    "CHUNK_SIZE_FRAC",
    "CheckEngineError",
    "Dialect",
    "GetDialectError",
    "GetTableError",
    "InsertItemsAsyncError",
    "InsertItemsError",
    "ParseEngineError",
    "TablenameMixin",
    "UpsertItemsAsyncError",
    "UpsertItemsError",
    "check_engine",
    "check_table_against_reflection",
    "columnwise_max",
    "columnwise_min",
    "create_engine",
    "ensure_engine",
    "ensure_tables_created",
    "ensure_tables_dropped",
    "ensure_tables_dropped_async",
    "get_chunk_size",
    "get_column_names",
    "get_columns",
    "get_dialect",
    "get_table",
    "get_table_name",
    "get_table_updated_column",
    "insert_items",
    "insert_items_async",
    "is_insert_item_pair",
    "is_mapped_class",
    "is_table_or_mapped_class",
    "is_upsert_item_pair",
    "is_upsert_item_pair",
    "mapped_class_to_dict",
    "parse_engine",
    "selectable_to_string",
    "serialize_engine",
    "upsert_items",
    "upsert_items_async",
    "yield_connection",
    "yield_connection_async",
    "yield_primary_key_columns",
]



from __future__ import annotations

import datetime as dt

from hypothesis import given
from hypothesis.strategies import booleans, dates, dictionaries, floats, lists
from pytest import raises

from utilities.hypothesis import int64s, text_ascii, zoned_datetimes
from utilities.orjson2 import deserialize2, serialize2
from utilities.sentinel import sentinel

_Object = bool | float | str | dt.date | dt.datetime
objects = (
    booleans()
    | floats(allow_nan=False, allow_infinity=False)
    | int64s()
    | text_ascii()
    | dates()
    | zoned_datetimes()
)


class TestSerializeAndDeserialize2:
    @given(obj=objects)
    def test_main(self, *, obj: _Object) -> None:
        result = deserialize2(serialize2(obj))
        assert result == obj

    @given(
        objects=lists(objects)
        | dictionaries(text_ascii(), objects)
        | lists(dictionaries(text_ascii(), objects))
        | dictionaries(text_ascii(), lists(objects))
    )
    def test_nested(
        self,
        *,
        objects: list[_Object]
        | dict[str, _Object]
        | list[dict[str, _Object]]
        | dict[str, list[_Object]],
    ) -> None:
        result = deserialize2(serialize2(objects))
        assert result == objects

    def test_arbitrary_objects(self) -> None:
        with raises(TypeError, match="Type is not JSON serializable: Sentinel"):
            _ = serialize2(sentinel)
        result = serialize2(sentinel, fallback=True)
        expected = b'"<sentinel>"'
        assert result == expected

    def test_error_serialize(self) -> None:
        with raises(TypeError, match="Type is not JSON serializable: Sentinel"):
            _ = serialize2(sentinel)



from __future__ import annotations

from collections.abc import Callable
from functools import wraps
from itertools import chain
from typing import Any, TypeVar, cast

from utilities.sys import trace

_F = TypeVar("_F", bound=Callable[..., Any])


def other_decorator(func: _F, /) -> _F:
    @wraps(func)
    def wrapped(*args: Any, **kwargs: Any) -> Any:
        return func(*args, **kwargs)

    return cast(_F, wrapped)


@trace
def func_decorated_first(
    a: int, b: int, /, *args: int, c: int = 0, **kwargs: int
) -> int:
    return func_decorated_second(2 * a, 2 * b, *args, c=2 * c, **kwargs)


@other_decorator
@trace(above=1)
def func_decorated_second(
    a: int, b: int, /, *args: int, c: int = 0, **kwargs: int
) -> int:
    return func_decorated_third(2 * a, 2 * b, *args, c=2 * c, **kwargs)


@trace()
@other_decorator
def func_decorated_third(
    a: int, b: int, /, *args: int, c: int = 0, **kwargs: int
) -> int:
    return func_decorated_fourth(2 * a, 2 * b, *args, c=2 * c, **kwargs)


@other_decorator
@trace(above=1)
@other_decorator
def func_decorated_fourth(
    a: int, b: int, /, *args: int, c: int = 0, **kwargs: int
) -> int:
    return func_decorated_fifth(2 * a, 2 * b, *args, c=2 * c, **kwargs)


@other_decorator
@other_decorator
@trace(above=2)
@other_decorator
@other_decorator
@other_decorator
def func_decorated_fifth(
    a: int, b: int, /, *args: int, c: int = 0, **kwargs: int
) -> int:
    result = sum(chain([a, b], args, [c], kwargs.values()))
    assert result > 0, f"Result ({result}) must be positive"
    return result



from __future__ import annotations

import asyncio
import time
from asyncio import Task, TaskGroup, sleep
from functools import partial
from re import search
from typing import TYPE_CHECKING, Any, Literal

from pytest import approx, fixture, raises

from tests.conftest import FLAKY
from utilities.functions import get_class_name
from utilities.iterables import one
from utilities.tracer import (
    NodeData,
    filter_failures,
    get_tracer_trees,
    set_tracer_trees,
    tracer,
)
from utilities.zoneinfo import HongKong

if TYPE_CHECKING:
    from collections.abc import Callable
    from pathlib import Path

    from utilities.treelib import Node


@fixture(autouse=True)
def set_tracer_tree_per_function() -> None:
    set_tracer_trees([])


class TestTracer:
    @FLAKY
    def test_sync(self) -> None:
        @tracer
        def outer(n: int, /) -> int:
            time.sleep(0.01)  # 0.01
            n = mid1(n + 1)  # 0.01
            return mid2(n + 1)  # 0.02

        @tracer
        def mid1(n: int, /) -> int:
            time.sleep(0.01)  # 0.01
            return n + 1

        @tracer
        def mid2(n: int, /) -> int:
            time.sleep(0.01)  # 0.01
            return inner(n + 1)  # e.01

        @tracer
        def inner(n: int, /) -> int:
            time.sleep(0.01)  # 0.01
            return n + 1

        assert outer(1) == 6
        tree = one(get_tracer_trees())
        root = tree[tree.root]
        self._check_node(root, outer, 0.04)
        node_mid1, node_mid2 = tree.children(root.identifier)
        self._check_node(node_mid1, mid1, 0.01)
        self._check_node(node_mid2, mid2, 0.02)
        assert len(tree.children(node_mid1.identifier)) == 0
        (node_inner,) = tree.children(node_mid2.identifier)
        self._check_node(node_inner, inner, 0.01)

    @FLAKY
    async def test_async(self) -> None:
        @tracer
        async def outer(n: int, /) -> int:
            await asyncio.sleep(0.01)  # 0.01
            n = await mid1(n + 1)  # 0.01
            return await mid2(n + 1)  # 0.02

        @tracer
        async def mid1(n: int, /) -> int:
            await asyncio.sleep(0.01)  # 0.01
            return n + 1

        @tracer
        async def mid2(n: int, /) -> int:
            await asyncio.sleep(0.01)  # 0.01
            return await inner(n + 1)  # 0.01

        @tracer
        async def inner(n: int, /) -> int:
            await asyncio.sleep(0.01)  # 0.01
            return n + 1

        assert await outer(1) == 6
        tree = one(get_tracer_trees())
        root = tree[tree.root]
        self._check_node(root, outer, 0.04)
        node_mid1, node_mid2 = tree.children(root.identifier)
        self._check_node(node_mid1, mid1, 0.01)
        self._check_node(node_mid2, mid2, 0.02)
        assert len(tree.children(node_mid1.identifier)) == 0
        (node_inner,) = tree.children(node_mid2.identifier)
        self._check_node(node_inner, inner, 0.01)

    def test_methods(self) -> None:
        class Example:
            @tracer
            def func(self, n: int, /) -> int:
                return n + 1

        assert Example().func(1) == 2
        tree = one(get_tracer_trees())
        root = tree[tree.root]
        assert (
            root.data.tag
            == "tests.test_tracer:TestTracer.test_methods.<locals>.Example.func"
        )

    @FLAKY
    def test_multiple_calls(self) -> None:
        @tracer
        def func(n: int, /) -> int:
            return n + 1

        assert func(1) == 2
        assert func(1) == 2
        trees = get_tracer_trees()
        assert len(trees) == 2

    def test_add_args_sync(self) -> None:
        @tracer(add_args=True)
        def func(n: int, /) -> int:
            return n + 1

        assert func(1) == 2
        self._check_add_args()

    async def test_add_args_async(self) -> None:
        @tracer(add_args=True)
        async def func(n: int, /) -> int:
            await asyncio.sleep(0.01)
            return n + 1

        assert await func(1) == 2
        self._check_add_args()

    def test_time_zone(self) -> None:
        @tracer(time_zone=HongKong)
        def func(n: int, /) -> int:
            return n + 1

        assert func(1) == 2
        tree = one(get_tracer_trees())
        data = tree[tree.root].data
        assert data.start_time.tzinfo is HongKong
        assert data.end_time is not None
        assert data.end_time.tzinfo is HongKong

    def test_pre_call_sync(self, *, tmp_path: Path) -> None:
        path = tmp_path.joinpath("log")

        @tracer(pre_call=partial(self.pre_call, path=path))
        def func(n: int, /) -> int:
            return n + 1

        assert func(1) == 2
        self._check_pre_call(path)

    async def test_pre_call_async(self, *, tmp_path: Path) -> None:
        path = tmp_path.joinpath("log")

        @tracer(pre_call=partial(self.pre_call, path=path))
        async def func(n: int, /) -> int:
            await asyncio.sleep(0.01)
            return n + 1

        assert await func(1) == 2
        self._check_pre_call(path)

    def test_suppress(self) -> None:
        @tracer(suppress=ValueError)
        def func() -> None:
            msg = "Always fails"
            raise ValueError(msg)

        with raises(ValueError, match="Always fails"):
            _ = func()
        self._check_error(func, outcome="suppressed")

    def test_post_error_sync(self, *, tmp_path: Path) -> None:
        path = tmp_path.joinpath("log")

        @tracer(post_error=partial(self.post_error, path=path))
        def func(n: int, /) -> int:
            if n >= 1:
                return n + 1
            msg = f"{n=} must be positive"
            raise ValueError(msg)

        with raises(ValueError, match="n=0 must be positive"):
            _ = func(0)
        self._check_post_error(path)

    async def test_post_error_async(self, *, tmp_path: Path) -> None:
        path = tmp_path.joinpath("log")

        @tracer(post_error=partial(self.post_error, path=path))
        async def func(n: int, /) -> int:
            await sleep(0.01)
            if n >= 1:
                return n + 1
            msg = f"{n=} must be positive"
            raise ValueError(msg)

        with raises(ValueError, match="n=0 must be positive"):
            _ = await func(0)
        self._check_post_error(path)

    def test_post_result_sync(self, *, tmp_path: Path) -> None:
        path = tmp_path.joinpath("log")

        @tracer(post_result=partial(self.post_result, path=path))
        def func(n: int, /) -> int:
            return n + 1

        assert func(1) == 2
        self._check_post_result(path)

    async def test_post_result_async(self, *, tmp_path: Path) -> None:
        path = tmp_path.joinpath("log")

        @tracer(post_result=partial(self.post_result, path=path))
        async def func(n: int, /) -> int:
            await asyncio.sleep(0.01)
            return n + 1

        assert await func(1) == 2
        self._check_post_result(path)

    def test_add_result_sync(self) -> None:
        @tracer(add_result=True)
        def func(n: int, /) -> int:
            return n + 1

        assert func(1) == 2
        self._check_add_result()

    async def test_add_result_async(self) -> None:
        @tracer(add_result=True)
        async def func(n: int, /) -> int:
            await asyncio.sleep(0.01)
            return n + 1

        assert await func(1) == 2
        self._check_add_result()

    def test_error_sync(self) -> None:
        @tracer
        def func() -> None:
            msg = "Always fails"
            raise ValueError(msg)

        with raises(ValueError, match="Always fails"):
            _ = func()
        self._check_error(func, outcome="failure")

    async def test_error_async(self) -> None:
        @tracer
        async def func() -> None:
            msg = "Always fails"
            raise ValueError(msg)

        with raises(ValueError, match="Always fails"):
            _ = await func()
        self._check_error(func, outcome="failure")

    def _check_node(
        self, node: Node, func: Callable[..., Any], duration: float, /
    ) -> None:
        tag = f"{func.__module__}:{func.__qualname__}"
        assert node.tag == tag
        data = node.data
        assert data.module == func.__module__
        assert data.qualname == func.__qualname__
        assert data.duration is not None
        assert data.duration.total_seconds() == approx(duration, abs=1.0)
        assert data.outcome == "success"

    def _check_add_args(self) -> None:
        tree = one(get_tracer_trees())
        data = tree[tree.root].data
        assert data.args == (1,)
        assert data.kwargs == {}

    def pre_call(self, _: NodeData[Any], n: int, /, *, path: Path) -> None:
        with path.open(mode="w") as fh:
            _ = fh.write(f"Calling with {n=}")  # pyright: ignore[reportAssignmentType]

    def _check_pre_call(self, path: Path, /) -> None:
        with path.open() as fh:
            assert fh.readlines() == ["Calling with n=1"]

    def post_error(self, data: NodeData[Any], /, *, path: Path) -> None:
        assert data.args is not None
        assert data.kwargs is not None
        assert data.end_time is not None
        assert data.outcome in {"failure", "suppressed"}
        assert data.error is not None
        with path.open(mode="w") as fh:
            _ = fh.write(
                f"Raised a {get_class_name(data.error)} with {data.args=}/{data.kwargs=}"
            )

    def _check_post_error(self, path: Path, /) -> None:
        with path.open() as fh:
            assert fh.readlines() == [
                "Raised a ValueError with data.args=(0,)/data.kwargs={}"
            ]

    def post_result(self, data: NodeData[Any], result: int, /, *, path: Path) -> None:
        assert data.end_time is not None
        assert data.outcome == "success"
        with path.open(mode="w") as fh:
            _ = fh.write(f"Result was {result}")

    def _check_post_result(self, path: Path, /) -> None:
        with path.open() as fh:
            assert fh.readlines() == ["Result was 2"]

    def _check_add_result(self) -> None:
        tree = one(get_tracer_trees())
        data = tree[tree.root].data
        assert data.result == 2

    def _check_error(
        self, func: Callable[..., Any], /, *, outcome: Literal["failure", "suppressed"]
    ) -> None:
        tree = one(get_tracer_trees())
        data = tree[tree.root].data
        assert data.outcome == outcome
        tag = f"{func.__module__}:{func.__qualname__}"
        timedelta = r"\d:\d{2}:\d{2}(?:\.\d{6})?"
        match outcome:
            case "failure":
                pattern = rf"^{tag} \(ValueError, {timedelta}\)$"
            case "suppressed":
                pattern = rf"^{tag} \({timedelta}\)$"
        assert search(pattern, data.desc)
        assert data.args is not None
        assert data.kwargs is not None
        assert data.end_time is not None
        assert data.outcome in {"failure", "suppressed"}
        assert isinstance(data.error, ValueError)


class TestFilterFailures:
    def test_sync(self) -> None:
        @tracer
        def outer(n: int, /) -> list[int]:
            return list(map(inner, range(n, -n - 1, -1)))

        @tracer
        def inner(n: int, /) -> int:
            if n >= 1:
                return n + 1
            msg = f"{n=} must be positive"
            raise ValueError(msg)

        with raises(ValueError, match="n=0 must be positive"):
            _ = outer(3)
        tree = one(get_tracer_trees())
        assert tree.size() == 5
        subtree = filter_failures(tree)
        assert subtree is not None
        assert subtree.size() == 2

    @FLAKY
    async def test_async(self) -> None:
        @tracer
        async def outer(n: int, /) -> list[int]:
            tasks: set[Task[int]] = set()
            async with TaskGroup() as tg:
                for i in range(n, -n - 1, -1):
                    tasks.add(tg.create_task(inner(i)))
            return [t.result() for t in tasks]

        @tracer
        async def inner(n: int, /) -> int:
            await asyncio.sleep(0.01)
            if n >= 1:
                return n + 1
            msg = f"{n=} must be positive"
            raise ValueError(msg)

        with raises(ExceptionGroup):
            _ = await outer(3)
        tree = one(get_tracer_trees())
        assert tree.size() == 8
        subtree = filter_failures(tree)
        assert subtree is not None
        assert subtree.size() == 5

    def test_no_failure(self) -> None:
        @tracer
        def func(n: int, /) -> int:
            return n + 1

        assert func(1) == 2
        tree = one(get_tracer_trees())
        result = filter_failures(tree)
        assert result is None

    @FLAKY
    async def test_list(self) -> None:
        @tracer
        async def outer(n: int, /) -> list[int]:
            tasks: set[Task[int]] = set()
            async with TaskGroup() as tg:
                for i in range(n, -n - 1, -1):
                    tasks.add(tg.create_task(inner(i)))
            return [t.result() for t in tasks]

        @tracer
        async def inner(n: int, /) -> int:
            await asyncio.sleep(0.01)
            if n >= 1:
                return n + 1
            msg = f"{n=} must be positive"
            raise ValueError(msg)

        assert await outer(-1) == []
        with raises(ExceptionGroup):
            _ = await outer(3)
        with raises(ExceptionGroup):
            _ = await outer(4)
        trees = get_tracer_trees()
        assert len(trees) == 3
        assert [t.size() for t in trees] == [1, 8, 10]
        subtrees = filter_failures(trees)
        assert len(subtrees) == 2
        assert [t.size() for t in subtrees] == [5, 6]



from __future__ import annotations

from itertools import chain

from utilities.sys import trace


@trace
def func_two_first(a: int, b: int, /, *args: int, c: int = 0, **kwargs: int) -> int:
    return func_two_second(2 * a, 2 * b, *args, c=2 * c, **kwargs)


@trace
def func_two_second(a: int, b: int, /, *args: int, c: int = 0, **kwargs: int) -> int:
    result = sum(chain([a, b], args, [c], kwargs.values()))
    assert result > 0, f"Result ({result}) must be positive"
    return result



from __future__ import annotations

import sys  # do use `from sys import ...`
from re import search
from typing import TYPE_CHECKING, Any, ClassVar, cast

from eventkit import Event
from hypothesis import HealthCheck, given, settings
from hypothesis.strategies import integers
from loguru import logger
from pytest import CaptureFixture

from utilities.eventkit import add_listener
from utilities.functions import identity
from utilities.loguru import HandlerConfiguration, LogLevel

if TYPE_CHECKING:
    from pytest import CaptureFixture


class TestAddListener:
    datetime: ClassVar[str] = r"^\d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2}\.\d{3} \| "

    @given(n=integers())
    @settings(suppress_health_check={HealthCheck.function_scoped_fixture})
    async def test_main(self, *, n: int) -> None:
        event = Event()
        _ = add_listener(event, identity)
        event.emit(n)

    @given(n=integers())
    @settings(suppress_health_check={HealthCheck.function_scoped_fixture})
    async def test_custom_error_handler(
        self, *, capsys: CaptureFixture, n: int
    ) -> None:
        def error(event: Event, exception: Exception, /) -> None:
            _ = (event, exception)
            print("Custom handler")  # noqa: T201

        event = Event()
        _ = add_listener(event, identity, error=error, _stdout=False, _loguru=False)
        event.emit(n, n)
        out = capsys.readouterr().out
        (line,) = out.splitlines()
        assert line == "Custom handler"

    @given(n=integers())
    @settings(suppress_health_check={HealthCheck.function_scoped_fixture})
    async def test_error_loguru(self, *, capsys: CaptureFixture, n: int) -> None:
        handler: HandlerConfiguration = {"sink": sys.stdout, "level": LogLevel.TRACE}
        _ = logger.configure(handlers=[cast(dict[str, Any], handler)])

        event = Event()
        _ = add_listener(event, identity, _stdout=False, _loguru=True)
        event.emit(n, n)
        out = capsys.readouterr().out
        (line1, line2, line3, *_, last) = out.splitlines()
        expected1 = r"ERROR    \| utilities\.eventkit:_add_listener_error:\d+ - Raised a TypeError whilst running 'Event':$"
        assert search(expected1, line1), line1
        pattern2 = r"^Event<Event, \[\[None, None, <function identity at .*>\]\]>$"
        assert search(pattern2, line2)
        assert line3 == "Traceback (most recent call last):"
        assert (
            last == "TypeError: identity() takes 1 positional argument but 2 were given"
        )

    @given(n=integers())
    @settings(suppress_health_check={HealthCheck.function_scoped_fixture})
    async def test_error_stdout(self, *, capsys: CaptureFixture, n: int) -> None:
        event = Event()
        _ = add_listener(event, identity, keep_ref=True, _stdout=True, _loguru=False)
        event.emit(n, n)
        out = capsys.readouterr().out
        (line1, line2, line3) = out.splitlines()
        assert line1 == "Raised a TypeError whilst running 'Event':"
        pattern2 = (
            r"^event=Event<Event, \[\[None, None, <function identity at .*>\]\]>$"
        )
        assert search(pattern2, line2)
        assert (
            line3
            == "exception=TypeError('identity() takes 1 positional argument but 2 were given')"
        )



from __future__ import annotations

import datetime as dt
import enum
from enum import auto
from operator import attrgetter
from re import search
from typing import TYPE_CHECKING, Any, TypeVar
from uuid import UUID

import sqlalchemy
from click import ParamType, argument, command, echo, option
from click.testing import CliRunner
from hypothesis import given
from hypothesis.strategies import (
    DataObject,
    SearchStrategy,
    booleans,
    data,
    dates,
    datetimes,
    floats,
    frozensets,
    integers,
    just,
    lists,
    sampled_from,
    times,
    uuids,
)
from pytest import mark, param

import utilities.click
import utilities.datetime
import utilities.types
from utilities.click import (
    Date,
    DirPath,
    Enum,
    ExistingDirPath,
    ExistingFilePath,
    FilePath,
    FrozenSetBools,
    FrozenSetDates,
    FrozenSetEnums,
    FrozenSetFloats,
    FrozenSetInts,
    FrozenSetMonths,
    FrozenSetStrs,
    FrozenSetUUIDs,
    ListBools,
    ListDates,
    ListEnums,
    ListFloats,
    ListInts,
    ListMonths,
    ListStrs,
    ListUUIDs,
    LocalDateTime,
    Time,
    Timedelta,
    ZonedDateTime,
)
from utilities.datetime import serialize_month
from utilities.hypothesis import (
    durations,
    months,
    sqlite_engines,
    text_ascii,
    timedeltas_2w,
)
from utilities.sqlalchemy import serialize_engine
from utilities.text import join_strs, strip_and_dedent
from utilities.whenever import (
    serialize_date,
    serialize_duration,
    serialize_local_datetime,
    serialize_time,
    serialize_timedelta,
    serialize_zoned_datetime,
)
from utilities.zoneinfo import UTC

if TYPE_CHECKING:
    from collections.abc import Callable, Iterable
    from pathlib import Path


_T = TypeVar("_T")


class TestFileAndDirPaths:
    def test_existing_dir_path(self, *, tmp_path: Path) -> None:
        @command()
        @argument("path", type=ExistingDirPath)
        def cli(*, path: Path) -> None:
            from pathlib import Path

            assert isinstance(path, Path)

        result = CliRunner().invoke(cli, [str(tmp_path)])
        assert result.exit_code == 0

        file_path = tmp_path.joinpath("file.txt")
        file_path.touch()
        result = CliRunner().invoke(cli, [str(file_path)])
        assert result.exit_code == 2
        assert search("is a file", result.stdout)

        non_existent = tmp_path.joinpath("non-existent")
        result = CliRunner().invoke(cli, [str(non_existent)])
        assert result.exit_code == 2
        assert search("does not exist", result.stdout)

    def test_existing_file_path(self, *, tmp_path: Path) -> None:
        @command()
        @argument("path", type=ExistingFilePath)
        def cli(*, path: Path) -> None:
            from pathlib import Path

            assert isinstance(path, Path)

        result = CliRunner().invoke(cli, [str(tmp_path)])
        assert result.exit_code == 2
        assert search("is a directory", result.stdout)

        file_path = tmp_path.joinpath("file.txt")
        file_path.touch()
        result = CliRunner().invoke(cli, [str(file_path)])
        assert result.exit_code == 0

        non_existent = tmp_path.joinpath("non-existent")
        result = CliRunner().invoke(cli, [str(non_existent)])
        assert result.exit_code == 2
        assert search("does not exist", result.stdout)

    def test_dir_path(self, *, tmp_path: Path) -> None:
        @command()
        @argument("path", type=DirPath)
        def cli(*, path: Path) -> None:
            from pathlib import Path

            assert isinstance(path, Path)

        result = CliRunner().invoke(cli, [str(tmp_path)])
        assert result.exit_code == 0

        file_path = tmp_path.joinpath("file.txt")
        file_path.touch()
        result = CliRunner().invoke(cli, [str(file_path)])
        assert result.exit_code == 2
        assert search("is a file", result.stdout)

        non_existent = tmp_path.joinpath("non-existent")
        result = CliRunner().invoke(cli, [str(non_existent)])
        assert result.exit_code == 0

    def test_file_path(self, *, tmp_path: Path) -> None:
        @command()
        @argument("path", type=FilePath)
        def cli(*, path: Path) -> None:
            from pathlib import Path

            assert isinstance(path, Path)

        result = CliRunner().invoke(cli, [str(tmp_path)])
        assert result.exit_code == 2
        assert search("is a directory", result.stdout)

        file_path = tmp_path.joinpath("file.txt")
        file_path.touch()
        result = CliRunner().invoke(cli, [str(file_path)])
        assert result.exit_code == 0

        non_existent = tmp_path.joinpath("non-existent")
        result = CliRunner().invoke(cli, [str(non_existent)])
        assert result.exit_code == 0


class _Truth(enum.Enum):
    true = auto()
    false = auto()


def _lift_serializer(serializer: Callable[[_T], str]) -> Callable[[Iterable[_T]], str]:
    def wrapped(values: Iterable[_T], /) -> str:
        return join_strs(map(serializer, values))

    return wrapped


class TestParameters:
    cases = (
        param(Date(), "DATE", dt.date, dates(), serialize_date, True),
        param(
            Enum(_Truth),
            "ENUM[_Truth]",
            _Truth,
            sampled_from(_Truth),
            attrgetter("name"),
            True,
        ),
        param(
            utilities.click.Duration(),
            "DURATION",
            utilities.types.Duration,
            durations(min_number=0, min_timedelta=dt.timedelta(0), two_way=True),
            serialize_duration,
            True,
        ),
        param(
            utilities.click.Engine(),
            "ENGINE",
            sqlalchemy.Engine,
            sqlite_engines(),
            serialize_engine,
            True,
        ),
        param(
            FrozenSetBools(),
            "FROZENSET[BOOL]",
            frozenset[bool],
            frozensets(booleans(), max_size=1),
            _lift_serializer(str),
            True,
        ),
        param(
            FrozenSetDates(),
            "FROZENSET[DATE]",
            frozenset[dt.date],
            frozensets(dates(), max_size=1),
            _lift_serializer(serialize_date),
            True,
        ),
        param(
            FrozenSetEnums(_Truth),
            "FROZENSET[ENUM[_Truth]]",
            frozenset[_Truth],
            frozensets(sampled_from(_Truth), max_size=1),
            _lift_serializer(attrgetter("name")),
            True,
        ),
        param(
            FrozenSetFloats(),
            "FROZENSET[FLOAT]",
            frozenset[float],
            frozensets(floats(0, 10), max_size=1),
            _lift_serializer(str),
            True,
        ),
        param(
            FrozenSetInts(),
            "FROZENSET[INT]",
            frozenset[int],
            frozensets(integers(0, 10), max_size=1),
            _lift_serializer(str),
            True,
        ),
        param(
            FrozenSetMonths(),
            "FROZENSET[MONTH]",
            frozenset[utilities.datetime.Month],
            frozensets(months(), max_size=1),
            _lift_serializer(serialize_month),
            True,
        ),
        param(
            FrozenSetStrs(),
            "FROZENSET[STRING]",
            frozenset[str],
            frozensets(text_ascii(), max_size=1),
            _lift_serializer(str),
            False,
        ),
        param(
            FrozenSetUUIDs(),
            "FROZENSET[UUID]",
            frozenset[UUID],
            frozensets(uuids(), max_size=1),
            _lift_serializer(str),
            True,
        ),
        param(
            ListBools(),
            "LIST[BOOL]",
            list[bool],
            lists(booleans()),
            _lift_serializer(str),
            True,
        ),
        param(
            ListDates(),
            "LIST[DATE]",
            list[dt.date],
            lists(dates()),
            _lift_serializer(serialize_date),
            True,
        ),
        param(
            ListEnums(_Truth),
            "LIST[ENUM[_Truth]]",
            list[_Truth],
            lists(sampled_from(_Truth)),
            _lift_serializer(attrgetter("name")),
            True,
        ),
        param(
            ListFloats(),
            "LIST[FLOAT]",
            list[float],
            lists(floats(0, 10)),
            _lift_serializer(str),
            True,
        ),
        param(
            ListInts(),
            "LIST[INT]",
            list[int],
            lists(integers(0, 10)),
            _lift_serializer(str),
            True,
        ),
        param(
            ListMonths(),
            "LIST[MONTH]",
            list[utilities.datetime.Month],
            lists(months()),
            _lift_serializer(serialize_month),
            True,
        ),
        param(
            ListStrs(),
            "LIST[STRING]",
            list[str],
            lists(text_ascii()),
            _lift_serializer(str),
            False,
        ),
        param(
            ListUUIDs(),
            "LIST[UUID]",
            list[UUID],
            lists(uuids()),
            _lift_serializer(str),
            True,
        ),
        param(
            LocalDateTime(),
            "LOCAL DATETIME",
            dt.datetime,
            datetimes(),
            serialize_local_datetime,
            True,
        ),
        param(
            utilities.click.Month(),
            "MONTH",
            utilities.datetime.Month,
            months(),
            serialize_month,
            True,
        ),
        param(Time(), "TIME", dt.time, times(), serialize_time, True),
        param(
            Timedelta(),
            "TIMEDELTA",
            dt.timedelta,
            timedeltas_2w(min_value=dt.timedelta(0)),
            serialize_timedelta,
            True,
        ),
        param(
            ZonedDateTime(),
            "ZONED DATETIME",
            dt.datetime,
            datetimes(timezones=just(UTC)),
            serialize_zoned_datetime,
            True,
        ),
    )

    @given(data=data())
    @mark.parametrize(
        ("param", "exp_repr", "cls", "strategy", "serialize", "failable"), cases
    )
    def test_argument(
        self,
        *,
        data: DataObject,
        param: ParamType,
        exp_repr: str,
        cls: Any,
        strategy: SearchStrategy[Any],
        serialize: Callable[[Any], str],
        failable: bool,
    ) -> None:
        assert repr(param) == exp_repr

        runner = CliRunner()

        @command()
        @argument("value", type=param)
        def cli(*, value: cls) -> None:
            echo(f"value = {serialize(value)}")

        value_str = serialize(data.draw(strategy))
        result = CliRunner().invoke(cli, [value_str])
        assert result.exit_code == 0
        assert result.stdout == f"value = {value_str}\n"

        result = runner.invoke(cli, ["error"])
        expected = 2 if failable else 0
        assert result.exit_code == expected

    @given(data=data())
    @mark.parametrize(
        ("param", "exp_repr", "cls", "strategy", "serialize", "failable"), cases
    )
    def test_option(
        self,
        *,
        data: DataObject,
        param: ParamType,
        exp_repr: str,
        cls: Any,
        strategy: SearchStrategy[Any],
        serialize: Callable[[Any], str],
        failable: bool,
    ) -> None:
        assert repr(param) == exp_repr

        value = data.draw(strategy)

        @command()
        @option("--value", type=param, default=value)
        def cli(*, value: cls) -> None:
            echo(f"value = {serialize(value)}")

        result = CliRunner().invoke(cli)
        assert result.exit_code == 0
        assert result.stdout == f"value = {serialize(value)}\n"

        _ = failable

    @mark.parametrize(
        "param", [param(ListEnums(_Truth)), param(FrozenSetEnums(_Truth))], ids=str
    )
    def test_error_list_and_frozensets_parse(self, *, param: ParamType) -> None:
        @command()
        @option("--value", type=param, default=0)
        def cli(*, value: list[_Truth] | frozenset[_Truth]) -> None:
            echo(f"value = {value}")

        result = CliRunner().invoke(cli)
        assert result.exit_code == 2
        assert search(
            "Invalid value for '--value': Unable to parse 0 of type <class 'int'>",
            result.stdout,
        )


class TestCLIHelp:
    @mark.parametrize(
        ("param", "expected"),
        [
            param(
                str,
                """
    Usage: cli [OPTIONS]

    Options:
      --value TEXT
      --help        Show this message and exit.
""",
            ),
            param(
                ListStrs(),
                """
    Usage: cli [OPTIONS]

    Options:
      --value [LIST[TEXT] SEP=,]
      --help                      Show this message and exit.
""",
            ),
            param(
                Enum(_Truth),
                """
    Usage: cli [OPTIONS]

    Options:
      --value [true,false]
      --help                Show this message and exit.
""",
            ),
            param(
                ListEnums(_Truth),
                """
    Usage: cli [OPTIONS]

    Options:
      --value [LIST[true,false] SEP=,]
      --help                          Show this message and exit.
""",
            ),
        ],
    )
    def test_main(self, *, param: Any, expected: str) -> None:
        @command()
        @option("--value", type=param)
        def cli(*, value: Any) -> None:
            echo(f"value = {value}")

        result = CliRunner().invoke(cli, ["--help"])
        assert result.exit_code == 0
        expected = strip_and_dedent(expected, trailing=True)
        assert result.stdout == expected



from __future__ import annotations

from re import escape
from typing import TYPE_CHECKING, Any

from pytest import mark, param, raises

from tests.test_sys_funcs.async_ import func_async
from tests.test_sys_funcs.decorated import (
    func_decorated_fifth,
    func_decorated_first,
    func_decorated_fourth,
    func_decorated_second,
    func_decorated_third,
)
from tests.test_sys_funcs.error import func_error_async, func_error_sync
from tests.test_sys_funcs.one import func_one
from tests.test_sys_funcs.two import func_two_first, func_two_second
from tests.test_sys_funcs.zero import func_zero
from utilities.iterables import one
from utilities.sentinel import sentinel
from utilities.sys import (
    VERSION_MAJOR_MINOR,
    _FrameInfo,
    _GetCallerOutput,
    get_caller,
    get_exc_trace_info,
)
from utilities.text import strip_and_dedent

if TYPE_CHECKING:
    from collections.abc import Callable


class TestGetCaller:
    @mark.parametrize(
        ("depth", "expected"),
        [param(1, "inner"), param(2, "outer"), param(3, "test_main")],
        ids=str,
    )
    def test_main(self, *, depth: int, expected: str) -> None:
        def outer() -> _GetCallerOutput:
            return inner()

        def inner() -> _GetCallerOutput:
            return get_caller(depth=depth)

        result = outer()
        assert result["module"] == "tests.test_sys"
        assert result["name"] == expected

    @mark.parametrize(
        ("depth", "expected"),
        [param(1, "inner"), param(2, "mid"), param(3, "outer"), param(4, "test_depth")],
        ids=str,
    )
    def test_depth(self, *, depth: int, expected: str) -> None:
        def outer() -> _GetCallerOutput:
            return mid()

        def mid() -> _GetCallerOutput:
            return inner()

        def inner() -> _GetCallerOutput:
            return get_caller(depth=depth)

        result = outer()
        assert result["module"] == "tests.test_sys"
        assert result["name"] == expected


class TestGetExcTraceInfo:
    def test_func_zero(self) -> None:
        result = func_zero(1, 2, 3, 4, c=5, d=6, e=7)
        assert result == 28
        try:
            _ = func_zero(1, 2, 3, 4, c=5, d=6, e=7, f=-result)
        except AssertionError:
            exc_info = get_exc_trace_info()
            assert exc_info.exc_type is AssertionError
            assert isinstance(exc_info.exc_value, AssertionError)
            assert exc_info.frames == []
        else:  # pragma: no cover
            msg = "Expected an assertion"
            raise AssertionError(msg)

    def test_func_one(self) -> None:
        result = func_one(1, 2, 3, 4, c=5, d=6, e=7)
        assert result == 28
        try:
            _ = func_one(1, 2, 3, 4, c=5, d=6, e=7, f=-result)
        except AssertionError:
            exc_info = get_exc_trace_info()
            assert exc_info.exc_type is AssertionError
            assert isinstance(exc_info.exc_value, AssertionError)
            frame = one(exc_info.frames)
            self._assert(
                frame, 1, 1, func_one, "one.py", 8, 11, self._assert_code_line, result
            )
        else:  # pragma: no cover
            msg = "Expected an assertion"
            raise AssertionError(msg)

    def test_func_two(self) -> None:
        result = func_two_first(1, 2, 3, 4, c=5, d=6, e=7)
        assert result == 36
        try:
            _ = func_two_first(1, 2, 3, 4, c=5, d=6, e=7, f=-result)
        except AssertionError:
            exc_info = get_exc_trace_info()
            assert exc_info.exc_type is AssertionError
            assert isinstance(exc_info.exc_value, AssertionError)
            expected = [
                (
                    8,
                    10,
                    func_two_first,
                    "return func_two_second(2 * a, 2 * b, *args, c=2 * c, **kwargs)",
                ),
                (13, 16, func_two_second, self._assert_code_line),
            ]
            for depth, (frame, (first_ln, ln, func, code_ln)) in enumerate(
                zip(exc_info.frames, expected, strict=True), start=1
            ):
                self._assert(
                    frame, depth, 2, func, "two.py", first_ln, ln, code_ln, result
                )
        else:  # pragma: no cover
            msg = "Expected an assertion"
            raise AssertionError(msg)

    def test_func_decorated(self) -> None:
        result = func_decorated_first(1, 2, 3, 4, c=5, d=6, e=7)
        assert result == 148
        try:
            _ = func_decorated_first(1, 2, 3, 4, c=5, d=6, e=7, f=-result)
        except AssertionError:
            exc_info = get_exc_trace_info()
            assert exc_info.exc_type is AssertionError
            assert isinstance(exc_info.exc_value, AssertionError)
            expected = [
                (
                    21,
                    25,
                    func_decorated_first,
                    "return func_decorated_second(2 * a, 2 * b, *args, c=2 * c, **kwargs)",
                ),
                (
                    28,
                    33,
                    func_decorated_second,
                    "return func_decorated_third(2 * a, 2 * b, *args, c=2 * c, **kwargs)",
                ),
                (
                    36,
                    41,
                    func_decorated_third,
                    "return func_decorated_fourth(2 * a, 2 * b, *args, c=2 * c, **kwargs)",
                ),
                (
                    44,
                    50,
                    func_decorated_fourth,
                    "return func_decorated_fifth(2 * a, 2 * b, *args, c=2 * c, **kwargs)",
                ),
                (53, 63, func_decorated_fifth, self._assert_code_line),
            ]
            for depth, (frame, (first_ln, ln, func, code_ln)) in enumerate(
                zip(exc_info.frames, expected, strict=True), start=1
            ):
                self._assert(
                    frame, depth, 5, func, "decorated.py", first_ln, ln, code_ln, result
                )
        else:  # pragma: no cover
            msg = "Expected an assertion"
            raise AssertionError(msg)

    async def test_func_async(self) -> None:
        result = await func_async(1, 2, 3, 4, c=5, d=6, e=7)
        assert result == 28
        try:
            _ = await func_async(1, 2, 3, 4, c=5, d=6, e=7, f=-result)
        except AssertionError:
            exc_info = get_exc_trace_info()
            assert exc_info.exc_type is AssertionError
            assert isinstance(exc_info.exc_value, AssertionError)
            frame = one(exc_info.frames)
            self._assert(
                frame,
                1,
                1,
                func_async,
                "async_.py",
                9,
                13,
                self._assert_code_line,
                result,
            )
        else:  # pragma: no cover
            msg = "Expected an assertion"
            raise AssertionError(msg)

    def test_pretty(self) -> None:
        result = func_two_first(1, 2, 3, 4, c=5, d=6, e=7)
        assert result == 36
        try:
            _ = func_two_first(1, 2, 3, 4, c=5, d=6, e=7, f=-result)
        except AssertionError:
            exc_info = get_exc_trace_info()
            result = exc_info.pretty(location=False)
            expected = strip_and_dedent("""
                Error running:

                  1. func_two_first
                  2. func_two_second
                  >> AssertionError: Result (0) must be positive

                Traced frames:

                  1/2. func_two_first
                    args[0] = 1
                    args[1] = 2
                    args[2] = 3
                    args[3] = 4
                    kwargs['c'] = 5
                    kwargs['d'] = 6
                    kwargs['e'] = 7
                    kwargs['f'] = -36
                    >> return func_two_second(2 * a, 2 * b, *args, c=2 * c, **kwargs)

                  2/2. func_two_second
                    args[0] = 2
                    args[1] = 4
                    args[2] = 3
                    args[3] = 4
                    kwargs['c'] = 10
                    kwargs['d'] = 6
                    kwargs['e'] = 7
                    kwargs['f'] = -36
                    >> assert result > 0, f"Result ({result}) must be positive"
                    >> AssertionError: Result (0) must be positive
            """)
            assert result == expected
        else:  # pragma: no cover
            msg = "Expected an assertion"
            raise AssertionError(msg)

    def test_error_sync(self) -> None:
        with raises(
            TypeError,
            match=escape(
                "func_error_sync() missing 1 required positional argument: 'b'"
            ),
        ):
            _ = func_error_sync(1)  # pyright: ignore[reportCallIssue]

    async def test_error_async(self) -> None:
        with raises(
            TypeError,
            match=escape(
                "func_error_async() takes 2 positional arguments but 3 were given"
            ),
        ):
            _ = await func_error_async(1, 2, 3)  # pyright: ignore[reportCallIssue]

    def _assert(
        self,
        frame: _FrameInfo,
        depth: int,
        max_depth: int,
        func: Callable[..., Any],
        filename: str,
        first_line_num: int,
        line_num: int,
        code_line: str,
        result: int,
        /,
    ) -> None:
        assert frame.depth == depth
        assert frame.max_depth == max_depth
        assert frame.func.__name__ == func.__name__
        assert frame.filename.parts[-2:] == ("test_sys_funcs", filename)
        assert frame.first_line_num == first_line_num
        assert frame.line_num == line_num
        assert frame.code_line == code_line
        assert frame.args == (2 ** (depth - 1), 2**depth, 3, 4)
        assert frame.kwargs == {"c": 5 * 2 ** (depth - 1), "d": 6, "e": 7, "f": -result}
        assert frame.result is sentinel
        assert isinstance(frame.error, AssertionError)

    @property
    def _assert_code_line(self) -> str:
        return 'assert result > 0, f"Result ({result}) must be positive"'


class TestVersionMajorMinor:
    def test_main(self) -> None:
        assert isinstance(VERSION_MAJOR_MINOR, tuple)
        expected = 2
        assert len(VERSION_MAJOR_MINOR) == expected



from __future__ import annotations

from collections.abc import Callable
from dataclasses import dataclass
from itertools import product
from logging import (
    Formatter,
    LogRecord,
    StreamHandler,
    basicConfig,
    getLevelNamesMapping,
    getLogger,
    setLogRecordFactory,
)
from logging.handlers import TimedRotatingFileHandler
from pathlib import Path
from re import search
from sys import stdout
from typing import TYPE_CHECKING, Any, ClassVar, Literal, assert_never, cast

from typing_extensions import override

from utilities.datetime import maybe_sub_pct_y
from utilities.git import get_repo_root

if TYPE_CHECKING:
    from zoneinfo import ZoneInfo

    from utilities.types import PathLike

try:
    from whenever import ZonedDateTime
except ModuleNotFoundError:  # pragma: no cover
    ZonedDateTime = None


LogLevel = Literal["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"]


def basic_config(
    *,
    format: str = "{asctime} | {name} | {levelname:8} | {message}",  # noqa: A002
) -> None:
    """Do the basic config."""
    basicConfig(
        format=format,
        datefmt=maybe_sub_pct_y("%Y-%m-%d %H:%M:%S"),
        style="{",
        level="DEBUG",
    )


def get_logging_level_number(level: LogLevel, /) -> int:
    """Get the logging level number."""
    mapping = getLevelNamesMapping()
    try:
        return mapping[level]
    except KeyError:
        raise GetLoggingLevelNumberError(level=level) from None


@dataclass(kw_only=True, slots=True)
class GetLoggingLevelNumberError(Exception):
    level: LogLevel

    @override
    def __str__(self) -> str:
        return f"Invalid logging level: {self.level!r}"


def _setup_logging_default_path() -> Path:
    return get_repo_root().joinpath(".logs")


def setup_logging(
    *,
    logger_name: str | None = None,
    console_level: LogLevel = "INFO",
    console_fmt: str = "❯ {zoned_datetime_str} | {name}:{funcName}:{lineno} | {message}",  # noqa: RUF001
    files_dir: PathLike | Callable[[], Path] | None = _setup_logging_default_path,
    files_when: str = "D",
    files_interval: int = 1,
    files_backup_count: int = 10,
    files_max_bytes: int = 10 * 1024**2,
    files_fmt: str = "{zoned_datetime_str} | {name}:{funcName}:{lineno} | {levelname:8} | {message}",
) -> None:
    """Set up logger."""
    # log record factory
    from tzlocal import get_localzone  # skipif-ci-and-windows

    class LogRecordNanoLocal(  # skipif-ci-and-windows
        _AdvancedLogRecord, time_zone=get_localzone()
    ): ...

    setLogRecordFactory(LogRecordNanoLocal)  # skipif-ci-and-windows

    console_fmt, files_fmt = [  # skipif-ci-and-windows
        f.replace("{zoned_datetime_str}", LogRecordNanoLocal.get_zoned_datetime_fmt())
        for f in [console_fmt, files_fmt]
    ]

    # logger
    logger = getLogger(name=logger_name)  # skipif-ci-and-windows
    logger.setLevel(get_logging_level_number("DEBUG"))  # skipif-ci-and-windows

    # formatter
    try:  # skipif-ci-and-windows
        from coloredlogs import DEFAULT_FIELD_STYLES, ColoredFormatter
    except ModuleNotFoundError:  # pragma: no cover
        console_formatter = Formatter(fmt=console_fmt, style="{")
        files_formatter = Formatter(fmt=files_fmt, style="{")
    else:  # skipif-ci-and-windows
        field_styles = DEFAULT_FIELD_STYLES | {
            "zoned_datetime_str": DEFAULT_FIELD_STYLES["asctime"]
        }
        console_formatter = ColoredFormatter(
            fmt=console_fmt, style="{", field_styles=field_styles
        )
        files_formatter = ColoredFormatter(
            fmt=files_fmt, style="{", field_styles=field_styles
        )
    plain_formatter = Formatter(fmt=files_fmt, style="{")  # skipif-ci-and-windows

    # console
    console_handler = StreamHandler(stream=stdout)  # skipif-ci-and-windows
    console_handler.setFormatter(console_formatter)  # skipif-ci-and-windows
    console_handler.setLevel(
        get_logging_level_number(console_level)
    )  # skipif-ci-and-windows
    logger.addHandler(console_handler)  # skipif-ci-and-windows

    # files
    match files_dir:  # skipif-ci-and-windows
        case None:
            directory = Path.cwd()
        case Path() | str():
            directory = Path(files_dir)
        case Callable():
            directory = files_dir()
        case _ as never:  # pyright: ignore[reportUnnecessaryComparison]
            assert_never(never)
    levels: list[LogLevel] = ["DEBUG", "INFO", "ERROR"]  # skipif-ci-and-windows
    for level, (subpath, formatter) in product(  # skipif-ci-and-windows
        levels, [(Path(), files_formatter), (Path("plain"), plain_formatter)]
    ):
        path = directory.joinpath(subpath, level.lower())
        path.parent.mkdir(parents=True, exist_ok=True)
        try:
            from concurrent_log_handler import ConcurrentTimedRotatingFileHandler
        except ModuleNotFoundError:  # pragma: no cover
            file_handler = TimedRotatingFileHandler(
                filename=str(path),
                when=files_when,
                interval=files_interval,
                backupCount=files_backup_count,
            )
        else:
            file_handler = ConcurrentTimedRotatingFileHandler(
                filename=str(path),
                when=files_when,
                interval=files_interval,
                backupCount=files_backup_count,
                maxBytes=files_max_bytes,
            )
        file_handler.setFormatter(formatter)
        file_handler.setLevel(level)
        logger.addHandler(file_handler)


class _AdvancedLogRecord(LogRecord):
    """Advanced log record."""

    time_zone: ClassVar[str] = NotImplemented

    @override
    def __init__(
        self,
        name: str,
        level: int,
        pathname: str,
        lineno: int,
        msg: object,
        args: Any,
        exc_info: Any,
        func: str | None = None,
        sinfo: str | None = None,
    ) -> None:
        self.zoned_datetime = self.get_now()  # skipif-ci-and-windows
        self.zoned_datetime_str = (  # skipif-ci-and-windows
            self.zoned_datetime.format_common_iso()
        )
        super().__init__(  # skipif-ci-and-windows
            name, level, pathname, lineno, msg, args, exc_info, func, sinfo
        )

    @override
    def __init_subclass__(cls, *, time_zone: ZoneInfo, **kwargs: Any) -> None:
        cls.time_zone = time_zone.key  # skipif-ci-and-windows
        super().__init_subclass__(**kwargs)  # skipif-ci-and-windows

    @override
    def getMessage(self) -> str:
        """Return the message for this LogRecord."""
        msg = str(self.msg)
        if self.args:
            try:
                return msg % self.args  # compability for 3rd party code
            except ValueError as error:
                if len(error.args) == 0:  # pragma: no cover
                    raise
                first = error.args[0]
                if search("unsupported format character", first):
                    return msg.format(*self.args)
                raise  # pragma: no cover
            except TypeError as error:  # pragma: no cover
                if len(error.args) == 0:
                    raise
                first = error.args[0]
                if search("not all arguments converted", first):
                    return msg.format(*self.args)
                raise
        return msg

    @classmethod
    def get_now(cls) -> Any:
        """Get the current zoned datetime."""
        return cast(Any, ZonedDateTime).now(cls.time_zone)  # skipif-ci-and-windows

    @classmethod
    def get_zoned_datetime_fmt(cls) -> str:
        """Get the zoned datetime format string."""
        length = len(cls.get_now().format_common_iso())  # skipif-ci-and-windows
        return f"{{zoned_datetime_str:{length}}}"  # skipif-ci-and-windows


__all__ = [
    "GetLoggingLevelNumberError",
    "LogLevel",
    "basic_config",
    "get_logging_level_number",
    "setup_logging",
]



from __future__ import annotations

from asyncio import sleep
from re import search
from typing import TYPE_CHECKING, Any, TypeVar, cast

from utilities.iterables import OneError, one
from utilities.text import EnsureStrError, ensure_str

if TYPE_CHECKING:
    from collections.abc import Coroutine

from asyncio import timeout
from collections.abc import AsyncIterable, Awaitable, Coroutine, Iterable
from typing import TYPE_CHECKING, TypeGuard

from utilities.datetime import duration_to_float

if TYPE_CHECKING:
    from asyncio import Timeout

    from utilities.types import Duration

_T = TypeVar("_T")
Coroutine1 = Coroutine[Any, Any, _T]
MaybeAwaitable = _T | Awaitable[_T]
MaybeCoroutine1 = _T | Coroutine1[_T]
_MaybeAsyncIterable = Iterable[_T] | AsyncIterable[_T]
_MaybeAwaitableMaybeAsyncIterable = MaybeAwaitable[_MaybeAsyncIterable[_T]]


async def is_awaitable(obj: Any, /) -> TypeGuard[Awaitable[Any]]:
    """Check if an object is awaitable."""
    try:
        await obj
    except TypeError:
        return False
    return True


async def sleep_dur(*, duration: Duration | None = None) -> None:
    """Sleep which accepts durations."""
    if duration is None:
        return
    await sleep(duration_to_float(duration))


def timeout_dur(*, duration: Duration | None = None) -> Timeout:
    """Timeout context manager which accepts durations."""
    delay = None if duration is None else duration_to_float(duration)
    return timeout(delay)


async def to_list(iterable: _MaybeAwaitableMaybeAsyncIterable[_T], /) -> list[_T]:
    """Reify an asynchronous iterable as a list."""
    value = await try_await(iterable)
    try:
        return [x async for x in cast(AsyncIterable[_T], value)]
    except TypeError:
        return list(cast(Iterable[_T], value))


async def try_await(obj: MaybeAwaitable[_T], /) -> _T:
    """Try await a value from an object."""
    try:
        return await cast(Awaitable[_T], obj)
    except TypeError as error:
        try:
            text = ensure_str(one(error.args))
        except (EnsureStrError, OneError):
            raise error from None
        if search("object .* can't be used in 'await' expression", text):
            return cast(_T, obj)
        raise


__all__ = [
    "Coroutine1",
    "MaybeAwaitable",
    "MaybeCoroutine1",
    "is_awaitable",
    "sleep_dur",
    "timeout_dur",
    "to_list",
]



from __future__ import annotations

from collections.abc import Callable, Iterable, Iterator
from dataclasses import dataclass, field
from functools import partial, wraps
from inspect import iscoroutinefunction, signature
from linecache import getline
from pathlib import Path
from sys import _getframe, exc_info, version_info
from textwrap import indent
from typing import TYPE_CHECKING, Any, TypedDict, TypeVar, cast, overload

from utilities.errors import ImpossibleCaseError
from utilities.functions import ensure_not_none, get_func_name
from utilities.sentinel import Sentinel, sentinel

if TYPE_CHECKING:
    from types import FrameType, TracebackType

    from utilities.sentinel import Sentinel


VERSION_MAJOR_MINOR = (version_info.major, version_info.minor)
_F = TypeVar("_F", bound=Callable[..., Any])
_TRACE_DATA = "_TRACE_DATA"


class _GetCallerOutput(TypedDict):
    module: str
    line_num: int
    name: str


def get_caller(*, depth: int = 2) -> _GetCallerOutput:
    """Get the calling function."""
    i = 0
    frame: FrameType | None = _getframe()  # pragma: no cover
    while (i < depth) and (frame.f_back is not None):
        i, frame = i + 1, frame.f_back
    return {
        "module": frame.f_globals["__name__"],
        "line_num": frame.f_lineno,
        "name": frame.f_code.co_name,
    }


@dataclass(kw_only=True)
class _GetExcTraceInfoOutput:
    """A collection of exception data."""

    exc_type: type[BaseException] | None = None
    exc_value: BaseException | None = None
    frames: list[_FrameInfo] = field(default_factory=list)

    def pretty(self, *, location: bool = True) -> str:
        """Pretty print the exception data."""
        return "\n".join(self._pretty_yield(location=location))

    def _pretty_yield(self, /, *, location: bool = True) -> Iterable[str]:
        """Yield the rows for pretty printing the exception."""
        from rich.pretty import pretty_repr

        if (self.exc_type is None) or (self.exc_value is None):  # pragma: no cover
            raise ImpossibleCaseError(case=[f"{self.exc_type=}", f"{self.exc_value=}"])
        error = f">> {self.exc_type.__name__}: {self.exc_value}"

        yield "Error running:"
        yield ""
        for frame in self.frames:
            yield indent(f"{frame.depth}. {get_func_name(frame.func)}", self._prefix1)
        yield indent(error, self._prefix1)
        yield ""
        yield "Traced frames:"
        for frame in self.frames:
            name, filename = get_func_name(frame.func), frame.filename
            yield ""
            desc = f"{name} ({filename}:{frame.first_line_num})" if location else name
            yield indent(f"{frame.depth}/{frame.max_depth}. {desc}", self._prefix1)
            for i, arg in enumerate(frame.args):
                yield indent(f"args[{i}] = {pretty_repr(arg)}", self._prefix2)
            for k, v in frame.kwargs.items():
                yield indent(f"kwargs[{k!r}] = {pretty_repr(v)}", self._prefix2)
            yield indent(f">> {frame.code_line}", self._prefix2)
            if location:  # pragma: no cover
                yield indent(f"   ({filename}:{frame.line_num})", self._prefix2)
            if frame.depth == frame.max_depth:
                yield indent(error, self._prefix2)

    @property
    def _prefix1(self) -> str:
        return 2 * " "

    @property
    def _prefix2(self) -> str:
        return 2 * self._prefix1


@dataclass(kw_only=True)
class _FrameInfo:
    """A collection of frame data."""

    depth: int
    max_depth: int
    func: Callable[..., Any]
    filename: Path
    first_line_num: int
    line_num: int
    code_line: str
    args: tuple[Any, ...] = field(default_factory=tuple)
    kwargs: dict[str, Any] = field(default_factory=dict)
    result: Any | Sentinel = sentinel
    error: Exception | None = None


def get_exc_trace_info() -> _GetExcTraceInfoOutput:
    """Get the exception information, extracting & merging trace data if it exists."""
    type_, value, traceback = exc_info()
    raw = list(_get_exc_trace_info_yield_raw(traceback=traceback))
    merged = list(_get_exc_trace_info_yield_merged(raw))
    frames = [
        _FrameInfo(
            depth=i,
            max_depth=len(merged),
            func=f.func,
            filename=f.filename,
            first_line_num=f.first_line_num,
            line_num=f.line_num,
            code_line=f.code_line,
            args=f.args,
            kwargs=f.kwargs,
            result=f.result,
            error=f.error,
        )
        for i, f in enumerate(merged[::-1], start=1)
    ]
    return _GetExcTraceInfoOutput(exc_type=type_, exc_value=value, frames=frames)


@dataclass(kw_only=True)
class _GetExcTraceInfoRaw:
    """A collection of raw frame data."""

    func_name: str
    filename: Path
    first_line_num: int
    line_num: int
    code_line: str
    trace: _TraceData | None = None


def _get_exc_trace_info_yield_raw(
    *, traceback: TracebackType | None = None
) -> Iterator[_GetExcTraceInfoRaw]:
    """Yield the raw frame info."""
    while traceback is not None:
        frame = traceback.tb_frame
        code = frame.f_code
        yield _GetExcTraceInfoRaw(
            func_name=code.co_name,
            filename=Path(code.co_filename),
            first_line_num=code.co_firstlineno,
            line_num=traceback.tb_lineno,
            code_line=getline(code.co_filename, traceback.tb_lineno).strip(),
            trace=frame.f_locals.get(_TRACE_DATA),
        )
        traceback = traceback.tb_next


@dataclass(kw_only=True)
class _GetExcTraceInfoMerged:
    """A collection of unnumbered frame data."""

    func: Callable[..., Any]
    filename: Path
    first_line_num: int
    line_num: int
    code_line: str
    args: tuple[Any, ...] = field(default_factory=tuple)
    kwargs: dict[str, Any] = field(default_factory=dict)
    result: Any | Sentinel = sentinel
    error: Exception | None = None


def _get_exc_trace_info_yield_merged(
    raw: Iterable[_GetExcTraceInfoRaw], /
) -> Iterator[_GetExcTraceInfoMerged]:
    """Yield the merged frame info."""
    raw_rev = list(reversed(list(raw)))
    while True:
        try:
            curr = raw_rev.pop(0)
        except IndexError:  # pragma: no cover
            return
        if len(raw_rev) == 0:
            return
        next_: _GetExcTraceInfoRaw | None = None
        while (len(raw_rev) >= 1) and ((next_ is None) or (next_.trace is None)):
            next_ = raw_rev.pop(0)
        next_ = ensure_not_none(next_)
        if next_.trace is None:
            return
        yield _GetExcTraceInfoMerged(
            func=next_.trace.func,
            filename=curr.filename,
            first_line_num=curr.first_line_num,
            line_num=curr.line_num,
            code_line=curr.code_line,
            args=next_.trace.args,
            kwargs=next_.trace.kwargs,
            result=next_.trace.result,
            error=next_.trace.error,
        )
        raw_rev = raw_rev[next_.trace.above :]


@dataclass(kw_only=True, slots=True)
class _TraceData:
    """A collection of tracing data."""

    func: Callable[..., Any]
    args: tuple[Any, ...] = field(default_factory=tuple)
    kwargs: dict[str, Any] = field(default_factory=dict)
    above: int = 0
    result: Any | Sentinel = sentinel
    error: Exception | None = None


@overload
def trace(func: _F, /, *, above: int = ...) -> _F: ...
@overload
def trace(func: None = None, /, *, above: int = ...) -> Callable[[_F], _F]: ...
def trace(func: _F | None = None, /, *, above: int = 0) -> _F | Callable[[_F], _F]:
    """Trace a function call."""
    if func is None:
        result = partial(trace, above=above)
        return cast(Callable[[_F], _F], result)

    if not iscoroutinefunction(func):

        @wraps(func)
        def trace_sync(*args: Any, **kwargs: Any) -> Any:
            try:
                trace_data = _trace_make_data(func, above, *args, **kwargs)
            except TypeError:
                return func(*args, **kwargs)
            try:
                result = trace_data.result = func(*args, **kwargs)
            except Exception as error:
                trace_data.error = error
                locals()[_TRACE_DATA] = trace_data
                raise
            locals()[_TRACE_DATA] = trace_data
            return result

        return trace_sync

    @wraps(func)
    async def log_call_async(*args: Any, **kwargs: Any) -> Any:
        try:
            trace_data = _trace_make_data(func, above, *args, **kwargs)
        except TypeError:
            return await func(*args, **kwargs)
        try:
            result = trace_data.result = await func(*args, **kwargs)
        except Exception as error:
            trace_data.error = error
            locals()[_TRACE_DATA] = trace_data
            raise
        locals()[_TRACE_DATA] = trace_data
        return result

    return cast(_F, log_call_async)


def _trace_make_data(
    func: Callable[..., Any], above: int = 0, *args: Any, **kwargs: Any
) -> _TraceData:
    """Make the initial trace data."""
    bound_args = signature(func).bind(*args, **kwargs)
    return _TraceData(
        func=func, args=bound_args.args, kwargs=bound_args.kwargs, above=above
    )


__all__ = ["VERSION_MAJOR_MINOR", "get_caller", "get_exc_trace_info", "trace"]



from __future__ import annotations

from dataclasses import dataclass
from textwrap import dedent
from typing import TYPE_CHECKING, Any, Literal, overload

from typing_extensions import override

from utilities.functions import get_class_name
from utilities.sentinel import SENTINEL_REPR
from utilities.types import EnsureClassError, ensure_class

if TYPE_CHECKING:
    from collections.abc import Iterable


@overload
def ensure_bytes(obj: Any, /, *, nullable: bool) -> bytes | None: ...
@overload
def ensure_bytes(obj: Any, /, *, nullable: Literal[False] = False) -> bytes: ...
def ensure_bytes(obj: Any, /, *, nullable: bool = False) -> bytes | None:
    """Ensure an object is a bytesean."""
    try:
        return ensure_class(obj, bytes, nullable=nullable)
    except EnsureClassError as error:
        raise EnsureBytesError(obj=error.obj, nullable=nullable) from None


@dataclass(kw_only=True, slots=True)
class EnsureBytesError(Exception):
    obj: Any
    nullable: bool

    @override
    def __str__(self) -> str:
        desc = " or None" if self.nullable else ""
        return f"Object {self.obj} must be a byte string{desc}; got {get_class_name(self.obj)} instead"


@overload
def ensure_str(obj: Any, /, *, nullable: bool) -> str | None: ...
@overload
def ensure_str(obj: Any, /, *, nullable: Literal[False] = False) -> str: ...
def ensure_str(obj: Any, /, *, nullable: bool = False) -> str | None:
    """Ensure an object is a string."""
    try:
        return ensure_class(obj, str, nullable=nullable)
    except EnsureClassError as error:
        raise EnsureStrError(obj=error.obj, nullable=nullable) from None


@dataclass(kw_only=True, slots=True)
class EnsureStrError(Exception):
    obj: Any
    nullable: bool

    @override
    def __str__(self) -> str:
        desc = " or None" if self.nullable else ""
        return f"Object {self.obj} must be a string{desc}; got {get_class_name(self.obj)} instead"


def join_strs(
    texts: Iterable[str], /, *, separator: str = ",", empty: str = SENTINEL_REPR
) -> str:
    """Join a collection of strings, with a special provision for the empty list."""
    texts = list(texts)
    if len(texts) >= 1:
        return separator.join(texts)
    return empty


def repr_encode(obj: Any, /) -> bytes:
    """Return the representation of the object encoded as bytes."""
    return repr(obj).encode()


def split_str(
    text: str, /, *, separator: str = ",", empty: str = SENTINEL_REPR
) -> list[str]:
    """Split a string, with a special provision for the empty string."""
    return [] if text == empty else text.split(separator)


def str_encode(obj: Any, /) -> bytes:
    """Return the string representation of the object encoded as bytes."""
    return str(obj).encode()


def strip_and_dedent(text: str, /, *, trailing: bool = False) -> str:
    """Strip and dedent a string."""
    result = dedent(text.strip("\n")).strip("\n")
    return f"{result}\n" if trailing else result


__all__ = [
    "EnsureBytesError",
    "EnsureStrError",
    "ensure_bytes",
    "ensure_str",
    "join_strs",
    "repr_encode",
    "split_str",
    "str_encode",
    "strip_and_dedent",
]



from __future__ import annotations

from dataclasses import fields, is_dataclass, replace
from itertools import chain
from typing import (
    TYPE_CHECKING,
    Any,
    ClassVar,
    TypeGuard,
    TypeVar,
    get_type_hints,
    runtime_checkable,
)

from typing_extensions import Protocol

from utilities.iterables import always_iterable
from utilities.sentinel import Sentinel

if TYPE_CHECKING:
    from collections.abc import Iterator

    from utilities.types import StrMapping


@runtime_checkable
class Dataclass(Protocol):
    """Protocol for `dataclass` classes."""

    __dataclass_fields__: ClassVar[dict[str, Any]]


def extend_non_sentinel(obj: _T, **kwargs: Any) -> _T:
    """Extend attributes on a dataclass, filtering out sentinel values."""
    return replace_non_sentinel(
        obj,
        **{
            k: list(chain(getattr(obj, k), always_iterable(v)))
            for k, v in kwargs.items()
            if not isinstance(v, Sentinel)
        },
    )


def get_dataclass_class(obj: Dataclass | type[Dataclass], /) -> type[Dataclass]:
    """Get the underlying dataclass, if possible."""
    if is_dataclass_class(obj):
        return obj
    if is_dataclass_instance(obj):
        return type(obj)
    msg = f"{obj=}"
    raise GetDataClassClassError(msg)


class GetDataClassClassError(Exception): ...


def get_dataclass_fields(
    cls: type[Dataclass],
    /,
    *,
    globalns: StrMapping | None = None,
    localns: StrMapping | None = None,
) -> StrMapping:
    """Get the fields of a dataclass."""
    return get_type_hints(
        cls,
        globalns=globals() if globalns is None else dict(globalns),
        localns=locals() if localns is None else dict(localns),
    )


def is_dataclass_class(obj: Any, /) -> TypeGuard[type[Dataclass]]:
    """Check if an object is a dataclass."""
    return isinstance(obj, type) and is_dataclass(obj)


def is_dataclass_instance(obj: Any, /) -> TypeGuard[Dataclass]:
    """Check if an object is an instance of a dataclass."""
    return (not isinstance(obj, type)) and is_dataclass(obj)


_T = TypeVar("_T", bound=Dataclass)


def replace_non_sentinel(obj: _T, **kwargs: Any) -> _T:
    """Replace attributes on a dataclass, filtering out sentinel values."""
    return replace(
        obj, **{k: v for k, v in kwargs.items() if not isinstance(v, Sentinel)}
    )


def yield_field_names(obj: Dataclass | type[Dataclass], /) -> Iterator[str]:
    """Yield the field names of a dataclass."""
    for field in fields(obj):
        yield field.name


__all__ = [
    "Dataclass",
    "GetDataClassClassError",
    "extend_non_sentinel",
    "get_dataclass_class",
    "get_dataclass_fields",
    "is_dataclass_class",
    "is_dataclass_instance",
    "replace_non_sentinel",
    "yield_field_names",
]
